# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import os
from pathlib import PureWindowsPath

from twisted.internet import error
from twisted.python.reflect import namedModule
from twisted.trial import unittest

from buildbot.process import remotetransfer
from buildbot.process.results import FAILURE
from buildbot.process.results import RETRY
from buildbot.process.results import SUCCESS
from buildbot.steps.source import bzr
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import ExpectCpdir
from buildbot.test.steps import ExpectDownloadFile
from buildbot.test.steps import ExpectRemoteRef
from buildbot.test.steps import ExpectRmdir
from buildbot.test.steps import ExpectShell
from buildbot.test.steps import ExpectStat
from buildbot.test.util import sourcesteps


class TestBzr(sourcesteps.SourceStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def test_mode_full(self):
        self.setup_step(
            bzr.Bzr(
                repourl='http://bzr.squid-cache.org/bzr/squid3/trunk', mode='full', method='fresh'
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['bzr', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.bzr', log_environ=True).exit(0),
            ExpectShell(workdir='wkdir', command=['bzr', 'clean-tree', '--force']).exit(0),
            ExpectShell(workdir='wkdir', command=['bzr', 'update']).exit(0),
            ExpectShell(
                workdir='wkdir', command=['bzr', 'version-info', '--custom', "--template='{revno}"]
            )
            .stdout('100')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS, state_string="update")
        self.expect_property('got_revision', '100', 'Bzr')
        return self.run_step()

    def test_mode_full_win32path(self):
        self.setup_step(
            bzr.Bzr(
                repourl='http://bzr.squid-cache.org/bzr/squid3/trunk', mode='full', method='fresh'
            )
        )
        self.build.path_module = namedModule('ntpath')
        self.build.path_cls = PureWindowsPath
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['bzr', '--version']).exit(0),
            ExpectStat(file=r'wkdir\.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file=r'wkdir\.bzr', log_environ=True).exit(0),
            ExpectShell(workdir='wkdir', command=['bzr', 'clean-tree', '--force']).exit(0),
            ExpectShell(workdir='wkdir', command=['bzr', 'update']).exit(0),
            ExpectShell(
                workdir='wkdir', command=['bzr', 'version-info', '--custom', "--template='{revno}"]
            )
            .stdout('100')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_full_timeout(self):
        self.setup_step(
            bzr.Bzr(
                repourl='http://bzr.squid-cache.org/bzr/squid3/trunk',
                mode='full',
                method='fresh',
                timeout=1,
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', timeout=1, command=['bzr', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.bzr', log_environ=True).exit(0),
            ExpectShell(workdir='wkdir', timeout=1, command=['bzr', 'clean-tree', '--force']).exit(
                0
            ),
            ExpectShell(workdir='wkdir', timeout=1, command=['bzr', 'update']).exit(0),
            ExpectShell(
                workdir='wkdir',
                timeout=1,
                command=['bzr', 'version-info', '--custom', "--template='{revno}"],
            )
            .stdout('100')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'Bzr')
        return self.run_step()

    def test_mode_full_revision(self):
        self.setup_step(
            bzr.Bzr(
                repourl='http://bzr.squid-cache.org/bzr/squid3/trunk', mode='full', method='fresh'
            ),
            args={"revision": '3730'},
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['bzr', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.bzr', log_environ=True).exit(0),
            ExpectShell(workdir='wkdir', command=['bzr', 'clean-tree', '--force']).exit(0),
            ExpectShell(workdir='wkdir', command=['bzr', 'update', '-r', '3730']).exit(0),
            ExpectShell(
                workdir='wkdir', command=['bzr', 'version-info', '--custom', "--template='{revno}"]
            )
            .stdout('100')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'Bzr')
        return self.run_step()

    def test_mode_full_clean(self):
        self.setup_step(
            bzr.Bzr(
                repourl='http://bzr.squid-cache.org/bzr/squid3/trunk', mode='full', method='clean'
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['bzr', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.bzr', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir', command=['bzr', 'clean-tree', '--ignored', '--force']
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['bzr', 'update']).exit(0),
            ExpectShell(
                workdir='wkdir', command=['bzr', 'version-info', '--custom', "--template='{revno}"]
            )
            .stdout('100')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'Bzr')
        return self.run_step()

    def test_mode_full_clean_patched(self):
        self.setup_step(
            bzr.Bzr(
                repourl='http://bzr.squid-cache.org/bzr/squid3/trunk', mode='full', method='clean'
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['bzr', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(0),
            # clean up the applied patch
            ExpectShell(
                workdir='wkdir', command=['bzr', 'clean-tree', '--ignored', '--force']
            ).exit(0),
            ExpectStat(file='wkdir/.bzr', log_environ=True).exit(0),
            # this clean is from 'mode=clean'
            ExpectShell(
                workdir='wkdir', command=['bzr', 'clean-tree', '--ignored', '--force']
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['bzr', 'update']).exit(0),
            ExpectShell(
                workdir='wkdir', command=['bzr', 'version-info', '--custom', "--template='{revno}"]
            )
            .stdout('100')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'Bzr')
        return self.run_step()

    def test_mode_full_clean_patch(self):
        self.setup_step(
            bzr.Bzr(
                repourl='http://bzr.squid-cache.org/bzr/squid3/trunk', mode='full', method='clean'
            ),
            patch=(1, 'patch'),
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['bzr', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.bzr', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir', command=['bzr', 'clean-tree', '--ignored', '--force']
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['bzr', 'update']).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.FileReader),
                workerdest='.buildbot-diff',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.FileReader),
                workerdest='.buildbot-patched',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'patch',
                    '-p1',
                    '--remove-empty-files',
                    '--force',
                    '--forward',
                    '-i',
                    '.buildbot-diff',
                ],
            ).exit(0),
            ExpectRmdir(dir='wkdir/.buildbot-diff', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir', command=['bzr', 'version-info', '--custom', "--template='{revno}"]
            )
            .stdout('100')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'Bzr')
        return self.run_step()

    def test_mode_full_clean_patch_worker_2_16(self):
        self.setup_build(worker_version={'*': '2.16'})
        self.setup_step(
            bzr.Bzr(
                repourl='http://bzr.squid-cache.org/bzr/squid3/trunk', mode='full', method='clean'
            ),
            patch=(1, 'patch'),
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['bzr', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.bzr', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir', command=['bzr', 'clean-tree', '--ignored', '--force']
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['bzr', 'update']).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.FileReader),
                slavedest='.buildbot-diff',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.FileReader),
                slavedest='.buildbot-patched',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'patch',
                    '-p1',
                    '--remove-empty-files',
                    '--force',
                    '--forward',
                    '-i',
                    '.buildbot-diff',
                ],
            ).exit(0),
            ExpectRmdir(dir='wkdir/.buildbot-diff', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir', command=['bzr', 'version-info', '--custom', "--template='{revno}"]
            )
            .stdout('100')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'Bzr')
        return self.run_step()

    def test_mode_full_clean_revision(self):
        self.setup_step(
            bzr.Bzr(
                repourl='http://bzr.squid-cache.org/bzr/squid3/trunk', mode='full', method='clean'
            ),
            args={"revision": '2345'},
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['bzr', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.bzr', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir', command=['bzr', 'clean-tree', '--ignored', '--force']
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['bzr', 'update', '-r', '2345']).exit(0),
            ExpectShell(
                workdir='wkdir', command=['bzr', 'version-info', '--custom', "--template='{revno}"]
            )
            .stdout('100')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'Bzr')
        return self.run_step()

    def test_mode_full_fresh(self):
        self.setup_step(
            bzr.Bzr(
                repourl='http://bzr.squid-cache.org/bzr/squid3/trunk', mode='full', method='fresh'
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['bzr', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.bzr', log_environ=True).exit(0),
            ExpectShell(workdir='wkdir', command=['bzr', 'clean-tree', '--force']).exit(0),
            ExpectShell(workdir='wkdir', command=['bzr', 'update']).exit(0),
            ExpectShell(
                workdir='wkdir', command=['bzr', 'version-info', '--custom', "--template='{revno}"]
            )
            .stdout('100')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'Bzr')
        return self.run_step()

    def test_mode_full_clobber(self):
        self.setup_step(
            bzr.Bzr(
                repourl='http://bzr.squid-cache.org/bzr/squid3/trunk', mode='full', method='clobber'
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['bzr', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['bzr', 'checkout', 'http://bzr.squid-cache.org/bzr/squid3/trunk', '.'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['bzr', 'version-info', '--custom', "--template='{revno}"]
            )
            .stdout('100')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'Bzr')
        return self.run_step()

    def test_mode_full_clobber_retry(self):
        self.setup_step(
            bzr.Bzr(
                repourl='http://bzr.squid-cache.org/bzr/squid3/trunk',
                mode='full',
                method='clobber',
                retry=(0, 2),
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['bzr', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['bzr', 'checkout', 'http://bzr.squid-cache.org/bzr/squid3/trunk', '.'],
            ).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['bzr', 'checkout', 'http://bzr.squid-cache.org/bzr/squid3/trunk', '.'],
            ).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['bzr', 'checkout', 'http://bzr.squid-cache.org/bzr/squid3/trunk', '.'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['bzr', 'version-info', '--custom', "--template='{revno}"]
            )
            .stdout('100')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'Bzr')
        return self.run_step()

    def test_mode_full_clobber_revision(self):
        self.setup_step(
            bzr.Bzr(
                repourl='http://bzr.squid-cache.org/bzr/squid3/trunk', mode='full', method='clobber'
            ),
            args={"revision": '3730'},
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['bzr', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'bzr',
                    'checkout',
                    'http://bzr.squid-cache.org/bzr/squid3/trunk',
                    '.',
                    '-r',
                    '3730',
                ],
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['bzr', 'version-info', '--custom', "--template='{revno}"]
            )
            .stdout('100')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'Bzr')
        return self.run_step()

    def test_mode_full_clobber_baseurl(self):
        self.setup_step(
            bzr.Bzr(
                baseURL='http://bzr.squid-cache.org/bzr/squid3',
                defaultBranch='trunk',
                mode='full',
                method='clobber',
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['bzr', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'bzr',
                    'checkout',
                    os.path.join('http://bzr.squid-cache.org/bzr/squid3', 'trunk'),
                    '.',
                ],
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['bzr', 'version-info', '--custom', "--template='{revno}"]
            )
            .stdout('100')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'Bzr')
        return self.run_step()

    def test_mode_full_clobber_baseurl_nodefault(self):
        self.setup_step(
            bzr.Bzr(
                baseURL='http://bzr.squid-cache.org/bzr/squid3',
                defaultBranch='trunk',
                mode='full',
                method='clobber',
            ),
            args={"branch": 'branches/SQUID_3_0'},
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['bzr', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'bzr',
                    'checkout',
                    os.path.join('http://bzr.squid-cache.org/bzr/squid3', 'branches/SQUID_3_0'),
                    '.',
                ],
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['bzr', 'version-info', '--custom', "--template='{revno}"]
            )
            .stdout('100')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'Bzr')
        return self.run_step()

    def test_mode_full_copy(self):
        self.setup_step(
            bzr.Bzr(
                repourl='http://bzr.squid-cache.org/bzr/squid3/trunk', mode='full', method='copy'
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['bzr', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='build', log_environ=True).exit(0),
            ExpectStat(file='source/.bzr', log_environ=True).exit(0),
            ExpectShell(workdir='source', command=['bzr', 'update']).exit(0),
            ExpectCpdir(fromdir='source', log_environ=True, todir='build').exit(0),
            ExpectShell(
                workdir='source', command=['bzr', 'version-info', '--custom', "--template='{revno}"]
            )
            .stdout('100')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'Bzr')
        return self.run_step()

    def test_mode_incremental(self):
        self.setup_step(
            bzr.Bzr(repourl='http://bzr.squid-cache.org/bzr/squid3/trunk', mode='incremental')
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['bzr', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.bzr', log_environ=True).exit(0),
            ExpectShell(workdir='wkdir', command=['bzr', 'update']).exit(0),
            ExpectShell(
                workdir='wkdir', command=['bzr', 'version-info', '--custom', "--template='{revno}"]
            )
            .stdout('100')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'Bzr')
        return self.run_step()

    def test_mode_incremental_revision(self):
        self.setup_step(
            bzr.Bzr(repourl='http://bzr.squid-cache.org/bzr/squid3/trunk', mode='incremental'),
            args={"revision": '9384'},
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['bzr', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.bzr', log_environ=True).exit(0),
            ExpectShell(workdir='wkdir', command=['bzr', 'update', '-r', '9384']).exit(0),
            ExpectShell(
                workdir='wkdir', command=['bzr', 'version-info', '--custom', "--template='{revno}"]
            )
            .stdout('100')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'Bzr')
        return self.run_step()

    def test_mode_incremental_no_existing_repo(self):
        self.setup_step(
            bzr.Bzr(repourl='http://bzr.squid-cache.org/bzr/squid3/trunk', mode='incremental')
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['bzr', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.bzr', log_environ=True).exit(1),
            ExpectShell(
                workdir='wkdir',
                command=['bzr', 'checkout', 'http://bzr.squid-cache.org/bzr/squid3/trunk', '.'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['bzr', 'version-info', '--custom', "--template='{revno}"]
            )
            .stdout('100\n')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100\n', 'Bzr')
        return self.run_step()

    def test_mode_incremental_retry(self):
        self.setup_step(
            bzr.Bzr(
                repourl='http://bzr.squid-cache.org/bzr/squid3/trunk',
                mode='incremental',
                retry=(0, 1),
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['bzr', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.bzr', log_environ=True).exit(1),
            ExpectShell(
                workdir='wkdir',
                command=['bzr', 'checkout', 'http://bzr.squid-cache.org/bzr/squid3/trunk', '.'],
            ).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['bzr', 'checkout', 'http://bzr.squid-cache.org/bzr/squid3/trunk', '.'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['bzr', 'version-info', '--custom', "--template='{revno}"]
            )
            .stdout('100\n')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100\n', 'Bzr')
        return self.run_step()

    def test_bad_revparse(self):
        self.setup_step(
            bzr.Bzr(repourl='http://bzr.squid-cache.org/bzr/squid3/trunk', mode='incremental')
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['bzr', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.bzr', log_environ=True).exit(1),
            ExpectShell(
                workdir='wkdir',
                command=['bzr', 'checkout', 'http://bzr.squid-cache.org/bzr/squid3/trunk', '.'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['bzr', 'version-info', '--custom', "--template='{revno}"]
            )
            .stdout('oiasdfj010laksjfd')
            .exit(0),
        )
        self.expect_outcome(result=FAILURE)
        return self.run_step()

    def test_bad_checkout(self):
        self.setup_step(
            bzr.Bzr(repourl='http://bzr.squid-cache.org/bzr/squid3/trunk', mode='incremental')
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['bzr', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.bzr', log_environ=True).exit(1),
            ExpectShell(
                workdir='wkdir',
                command=['bzr', 'checkout', 'http://bzr.squid-cache.org/bzr/squid3/trunk', '.'],
            )
            .stderr('failed\n')
            .exit(128),
        )
        self.expect_outcome(result=FAILURE)
        return self.run_step()

    def test_worker_connection_lost(self):
        self.setup_step(
            bzr.Bzr(
                repourl='http://bzr.squid-cache.org/bzr/squid3/trunk', mode='full', method='fresh'
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['bzr', '--version']).error(error.ConnectionLost())
        )
        self.expect_outcome(result=RETRY, state_string="update (retry)")
        return self.run_step()



#!/usr/bin/env python

try:
    from buildbot_pkg import setup_www_plugin
except ImportError:
    import sys

    print(
        'Please install buildbot_pkg module in order to install that '
        'package, or use the pre-build .whl modules available on pypi',
        file=sys.stderr,
    )
    sys.exit(1)

setup_www_plugin(
    name='buildbot-nestedexample',
    description='"An example of a custom nested parameter"',
    author='Ion Alberdi',
    author_email='ialberdi@intel.com',
    url='http://buildbot.net/',
    version='0.0.1',
    packages=['buildbot_nestedexample'],
    install_requires=['klein'],
    package_data={'': ['VERSION', 'static/*']},
    entry_points="""
        [buildbot.www]
        nestedexample = buildbot_nestedexample:ep
    """,
    classifiers=['License :: OSI Approved :: GNU General Public License v2 (GPLv2)'],
)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Member

from __future__ import annotations

from typing import TYPE_CHECKING

from twisted.internet import defer

from buildbot.data import base
from buildbot.data import types
from buildbot.db.logs import LogSlugExistsError
from buildbot.util import identifiers

if TYPE_CHECKING:
    from buildbot.db.logs import LogModel
    from buildbot.util.twisted import InlineCallbacksType


class EndpointMixin:
    def db2data(self, model: LogModel):
        data = {
            'logid': model.id,
            'name': model.name,
            'slug': model.slug,
            'stepid': model.stepid,
            'complete': model.complete,
            'num_lines': model.num_lines,
            'type': model.type,
        }
        return defer.succeed(data)


class LogEndpoint(EndpointMixin, base.BuildNestingMixin, base.Endpoint):
    kind = base.EndpointKind.SINGLE
    pathPatterns = [
        "/logs/n:logid",
        "/steps/n:stepid/logs/i:log_slug",
        "/builds/n:buildid/steps/i:step_name/logs/i:log_slug",
        "/builds/n:buildid/steps/n:step_number/logs/i:log_slug",
        "/builders/n:builderid/builds/n:build_number/steps/i:step_name/logs/i:log_slug",
        "/builders/n:builderid/builds/n:build_number/steps/n:step_number/logs/i:log_slug",
        "/builders/s:buildername/builds/n:build_number/steps/i:step_name/logs/i:log_slug",
        "/builders/s:buildername/builds/n:build_number/steps/n:step_number/logs/i:log_slug",
    ]

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        if 'logid' in kwargs:
            dbdict = yield self.master.db.logs.getLog(kwargs['logid'])
            return (yield self.db2data(dbdict)) if dbdict else None

        retriever = base.NestedBuildDataRetriever(self.master, kwargs)
        step_dict = yield retriever.get_step_dict()
        if step_dict is None:
            return None

        dbdict = yield self.master.db.logs.getLogBySlug(step_dict.id, kwargs.get('log_slug'))
        return (yield self.db2data(dbdict)) if dbdict else None


class LogsEndpoint(EndpointMixin, base.BuildNestingMixin, base.Endpoint):
    kind = base.EndpointKind.COLLECTION
    pathPatterns = [
        "/steps/n:stepid/logs",
        "/builds/n:buildid/steps/i:step_name/logs",
        "/builds/n:buildid/steps/n:step_number/logs",
        "/builders/n:builderid/builds/n:build_number/steps/i:step_name/logs",
        "/builders/n:builderid/builds/n:build_number/steps/n:step_number/logs",
        "/builders/s:buildername/builds/n:build_number/steps/i:step_name/logs",
        "/builders/s:buildername/builds/n:build_number/steps/n:step_number/logs",
    ]

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        retriever = base.NestedBuildDataRetriever(self.master, kwargs)
        step_dict = yield retriever.get_step_dict()
        if step_dict is None:
            return []
        logs = yield self.master.db.logs.getLogs(stepid=step_dict.id)
        results = []
        for dbdict in logs:
            results.append((yield self.db2data(dbdict)))
        return results


class Log(base.ResourceType):
    name = "log"
    plural = "logs"
    endpoints = [LogEndpoint, LogsEndpoint]
    eventPathPatterns = [
        "/logs/:logid",
        "/steps/:stepid/logs/:slug",
    ]

    class EntityType(types.Entity):
        logid = types.Integer()
        name = types.String()
        slug = types.Identifier(50)
        stepid = types.Integer()
        complete = types.Boolean()
        num_lines = types.Integer()
        type = types.Identifier(1)

    entityType = EntityType(name)

    @defer.inlineCallbacks
    def generateEvent(self, _id, event):
        # get the build and munge the result for the notification
        build = yield self.master.data.get(('logs', str(_id)))
        self.produceEvent(build, event)

    @base.updateMethod
    @defer.inlineCallbacks
    def addLog(self, stepid: int, name: str, type: int) -> InlineCallbacksType[int]:
        slug = identifiers.forceIdentifier(50, name)
        while True:
            try:
                logid = yield self.master.db.logs.addLog(
                    stepid=stepid, name=name, slug=slug, type=type
                )
            except LogSlugExistsError:
                slug = identifiers.incrementIdentifier(50, slug)
                continue
            self.generateEvent(logid, "new")
            return logid

    @base.updateMethod
    @defer.inlineCallbacks
    def appendLog(self, logid: int, content: str) -> InlineCallbacksType[None]:
        res = yield self.master.db.logs.appendLog(logid=logid, content=content)
        self.generateEvent(logid, "append")
        return res

    @base.updateMethod
    @defer.inlineCallbacks
    def finishLog(self, logid: int) -> InlineCallbacksType[None]:
        res = yield self.master.db.logs.finishLog(logid=logid)
        self.generateEvent(logid, "finished")
        return res

    @base.updateMethod
    def compressLog(self, logid: int) -> defer.Deferred[int]:
        return self.master.db.logs.compressLog(logid=logid)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING
from typing import Any
from typing import Awaitable
from typing import Callable

from buildbot.db import base
from buildbot.util.twisted import async_to_deferred

if TYPE_CHECKING:
    import sqlalchemy as sa
    from twisted.internet import defer

    from buildbot.data.resultspec import ResultSpec


@dataclass
class CodebaseCommitModel:
    id: int
    codebaseid: int
    author: str
    committer: str | None
    comments: str
    when_timestamp: int
    revision: str
    parent_commitid: int | None


UNKNOWN_COMMIT_ID = -1


@dataclass
class CommonCommitInfo:
    common_commit_id: int
    to1_commit_ids: list[int]
    to2_commit_ids: list[int]


class CodebaseCommitCache:
    def __init__(self) -> None:
        self._parents: dict[int, int] = {}

    def add_parent(self, id: int, parent_id: int) -> None:
        self._parents[id] = parent_id

    def get_parent(self, id: int) -> int:
        return self._parents.get(id, UNKNOWN_COMMIT_ID)

    async def get_parent_with_fallback(
        self, id: int, get_parent_fallback: Callable[[int], Awaitable[int]], default: int | None
    ) -> int | None:
        parent_id: int | None = self.get_parent(id)
        if parent_id == UNKNOWN_COMMIT_ID:
            parent_id = await get_parent_fallback(id)
            if parent_id == UNKNOWN_COMMIT_ID:
                parent_id = default
            else:
                self.add_parent(id, parent_id)
        return parent_id

    async def first_common_parent_with_ranges(
        self,
        id1: int,
        id2: int,
        get_parent_fallback: Callable[[int], Awaitable[int]],
        depth: int = 100,
    ) -> CommonCommitInfo | None:
        """
        Evaluates the commit graph and finds first parent of both id1 and id2 commits.

        The function returns a tuple that consists of:
        - the parent commit id
        - list of commit IDs from the parent commit to id1 (including parent commit and id1)
        - list of commit IDs from the parent commit to id2 (including parent commit and id2)

        If no parent is found, returns None
        """
        if id1 == id2:
            parent1 = await self.get_parent_with_fallback(
                id1, get_parent_fallback, UNKNOWN_COMMIT_ID
            )
            if parent1 == UNKNOWN_COMMIT_ID:
                return None
            return CommonCommitInfo(id1, [id1], [id1])

        parent1 = id1
        parent2 = id2
        known1 = [id1]
        known2 = [id2]

        for i in range(depth):
            parent1 = await self.get_parent_with_fallback(parent1, get_parent_fallback, None)
            if parent1 is None:
                break
            known1.append(parent1)

        if parent2 in known1:
            return CommonCommitInfo(parent2, known1[known1.index(parent2) + 1 :: -1], known2)

        for i in range(depth):
            parent2_new = await self.get_parent_with_fallback(parent2, get_parent_fallback, None)
            if parent2_new is None:
                break
            parent2 = parent2_new
            known2.append(parent2)
            if parent2 in known1:
                return CommonCommitInfo(parent2, known1[known1.index(parent2) :: -1], known2[::-1])

        return None


class CodebaseCommitsConnectorComponent(base.DBConnectorComponent):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._cache = CodebaseCommitCache()

    @async_to_deferred
    async def get_commit_by_revision(
        self, *, codebaseid: int, revision: str
    ) -> CodebaseCommitModel | None:
        def thd(conn: sa.engine.Connection) -> CodebaseCommitModel | None:
            tbl = self.db.model.codebase_commits
            q = tbl.select().where(
                (tbl.c.codebaseid == codebaseid) & (tbl.c.revision == revision),
            )
            res = conn.execute(q)
            row = res.fetchone()

            rv = None
            if row:
                rv = self._model_from_row(row)
            res.close()
            return rv

        commit = await self.db.pool.do(thd)
        if commit is not None:
            self._cache.add_parent(commit.id, commit.parent_commitid)
        return commit

    @async_to_deferred
    async def _get_commit_parent_commitid(self, id: int) -> int:
        def thd(conn: sa.engine.Connection) -> int | None:
            q = self.db.model.codebase_commits.select().where(
                self.db.model.codebase_commits.c.id == id,
            )
            res = conn.execute(q)
            row = res.fetchone()

            rv = None
            if row:
                rv = row.parent_commitid
            else:
                rv = UNKNOWN_COMMIT_ID
            res.close()
            return rv

        return await self.db.pool.do(thd)

    @async_to_deferred
    async def get_first_common_commit_with_ranges(
        self, first_commitid: int, last_commitid: int, depth: int = 100
    ) -> CommonCommitInfo | None:
        return await self._cache.first_common_parent_with_ranges(
            first_commitid, last_commitid, self._get_commit_parent_commitid, depth=depth
        )

    def get_commit(self, id: int) -> defer.Deferred[CodebaseCommitModel | None]:
        def thd(conn: sa.engine.Connection) -> CodebaseCommitModel | None:
            tbl = self.db.model.codebase_commits
            q = tbl.select().where(tbl.c.id == id)
            res = conn.execute(q)

            row = res.fetchone()
            rv = self._model_from_row(row) if row else None
            res.close()
            return rv

        return self.db.pool.do(thd)

    def get_commits_by_id(
        self, commit_ids: list[int]
    ) -> defer.Deferred[list[CodebaseCommitModel | None]]:
        def thd(conn: sa.engine.Connection) -> list[CodebaseCommitModel | None]:
            tbl = self.db.model.codebase_commits
            q = tbl.select().where(tbl.c.id.in_(commit_ids))
            res = conn.execute(q)
            commits = {row.id: self._model_from_row(row) for row in res.fetchall()}
            res.close()
            return [commits.get(id, None) for id in commit_ids]

        return self.db.pool.do(thd)

    def get_commits(
        self, *, codebaseid: int, result_spec: ResultSpec | None = None
    ) -> defer.Deferred[list[CodebaseCommitModel]]:
        def thd(conn: sa.engine.Connection) -> list[CodebaseCommitModel]:
            tbl = self.db.model.codebase_commits
            q = tbl.select().where(tbl.c.codebaseid == codebaseid)

            if result_spec is not None:
                return result_spec.thd_execute(conn, q, lambda row: self._model_from_row(row))

            res = conn.execute(q)
            rv = [self._model_from_row(row) for row in res.fetchall()]
            res.close()
            return rv

        return self.db.pool.do(thd)

    def _model_from_row(self, row: sa.engine.Row) -> CodebaseCommitModel:
        return CodebaseCommitModel(
            id=row.id,
            codebaseid=row.codebaseid,
            author=row.author,
            committer=row.committer,
            comments=row.comments,
            when_timestamp=row.when_timestamp,
            revision=row.revision,
            parent_commitid=row.parent_commitid,
        )

    @async_to_deferred
    async def add_commit(
        self,
        *,
        codebaseid: int,
        author: str,
        committer: str | None = None,
        files: list[str] | None = None,
        comments: str,
        when_timestamp: int,
        revision: str,
        parent_commitid: int | None = None,
    ) -> int:
        self.checkLength(self.db.model.codebase_commits.c.author, author)
        self.checkLength(self.db.model.codebase_commits.c.committer, committer)
        self.checkLength(self.db.model.codebase_commits.c.revision, revision)

        # TODO: handle files

        def thd(conn: sa.engine.Connection) -> int:
            q = self.db.model.codebase_commits.insert()
            r = conn.execute(
                q.values(
                    codebaseid=codebaseid,
                    author=author,
                    committer=committer,
                    comments=comments,
                    when_timestamp=when_timestamp,
                    revision=revision,
                    parent_commitid=parent_commitid,
                )
            )
            conn.commit()
            got_id = r.inserted_primary_key[0]
            r.close()

            return got_id

        return await self.db.pool.do_with_transaction(thd)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import os
import sys
from typing import Callable
from typing import Mapping
from typing import NoReturn
from typing import cast

from twisted.python import log
from twisted.python import usage
from twisted.trial import unittest

from buildbot_worker.scripts import runner
from buildbot_worker.test.util import misc

try:
    from unittest import mock
except ImportError:
    from unittest import mock


class OptionsMixin:
    def assertOptions(self, opts: Mapping, exp: Mapping) -> None:
        assert isinstance(self, unittest.TestCase)
        got = {k: opts[k] for k in exp}
        if got != exp:
            msg = []
            for k in exp:
                if opts[k] != exp[k]:
                    msg.append(f" {k}: expected {exp[k]!r}, got {opts[k]!r}")
            self.fail("did not get expected options\n" + ("\n".join(msg)))


class BaseDirTestsMixin:
    """
    Common tests for Options classes with 'basedir' parameter
    """

    GETCWD_PATH = "test-dir"
    ABSPATH_PREFIX = "test-prefix-"
    MY_BASEDIR = "my-basedir"

    # the options class to instantiate for test cases
    options_class: type[usage.Options] | None = None

    def setUp(self) -> None:
        assert isinstance(self, unittest.TestCase)
        self.patch(os, "getcwd", lambda: self.GETCWD_PATH)
        self.patch(os.path, "abspath", lambda path: self.ABSPATH_PREFIX + path)

    def parse(self, *args: str) -> usage.Options:
        assert self.options_class is not None

        opts = self.options_class()
        opts.parseOptions(args)
        return opts

    def test_defaults(self) -> None:
        assert isinstance(self, unittest.TestCase)
        opts = self.parse()
        self.assertEqual(
            opts["basedir"], self.ABSPATH_PREFIX + self.GETCWD_PATH, "unexpected basedir path"
        )

    def test_basedir_arg(self) -> None:
        assert isinstance(self, unittest.TestCase)
        opts = self.parse(self.MY_BASEDIR)
        self.assertEqual(
            opts["basedir"], self.ABSPATH_PREFIX + self.MY_BASEDIR, "unexpected basedir path"
        )

    def test_too_many_args(self) -> None:
        assert isinstance(self, unittest.TestCase)
        with self.assertRaisesRegex(usage.UsageError, "I wasn't expecting so many arguments"):
            self.parse("arg1", "arg2")


class TestMakerBase(BaseDirTestsMixin, unittest.TestCase):
    """
    Test buildbot_worker.scripts.runner.MakerBase class.
    """

    options_class = runner.MakerBase


class TestStopOptions(BaseDirTestsMixin, unittest.TestCase):
    """
    Test buildbot_worker.scripts.runner.StopOptions class.
    """

    options_class = runner.StopOptions

    def test_synopsis(self) -> None:
        opts = runner.StopOptions()
        self.assertIn('buildbot-worker stop', opts.getSynopsis())


class TestStartOptions(OptionsMixin, BaseDirTestsMixin, unittest.TestCase):
    """
    Test buildbot_worker.scripts.runner.StartOptions class.
    """

    options_class = runner.StartOptions

    def test_synopsis(self) -> None:
        opts = runner.StartOptions()
        self.assertIn('buildbot-worker start', opts.getSynopsis())

    def test_all_args(self) -> None:
        opts = self.parse("--quiet", "--nodaemon", self.MY_BASEDIR)
        self.assertOptions(
            opts,
            {"quiet": True, "nodaemon": True, "basedir": self.ABSPATH_PREFIX + self.MY_BASEDIR},
        )


class TestRestartOptions(OptionsMixin, BaseDirTestsMixin, unittest.TestCase):
    """
    Test buildbot_worker.scripts.runner.RestartOptions class.
    """

    options_class = runner.RestartOptions

    def test_synopsis(self) -> None:
        opts = runner.RestartOptions()
        self.assertIn('buildbot-worker restart', opts.getSynopsis())

    def test_all_args(self) -> None:
        opts = self.parse("--quiet", "--nodaemon", self.MY_BASEDIR)
        self.assertOptions(
            opts,
            {"quiet": True, "nodaemon": True, "basedir": self.ABSPATH_PREFIX + self.MY_BASEDIR},
        )


class TestCreateWorkerOptions(OptionsMixin, unittest.TestCase):
    """
    Test buildbot_worker.scripts.runner.CreateWorkerOptions class.
    """

    req_args = ["bdir", "mstr:5678", "name", "pswd"]

    def parse(self, *args: str) -> runner.CreateWorkerOptions:
        opts = runner.CreateWorkerOptions()
        opts.parseOptions(args)
        return opts

    def test_defaults(self) -> None:
        with self.assertRaisesRegex(usage.UsageError, "incorrect number of arguments"):
            self.parse()

    def test_synopsis(self) -> None:
        opts = runner.CreateWorkerOptions()
        self.assertIn('buildbot-worker create-worker', opts.getSynopsis())

    def test_min_args(self) -> None:
        # patch runner.MakerBase.postOptions() so that 'basedir'
        # argument will not be converted to absolute path
        self.patch(runner.MakerBase, "postOptions", mock.Mock())

        self.assertOptions(
            self.parse(*self.req_args),
            {"basedir": "bdir", "host": "mstr", "port": 5678, "name": "name", "passwd": "pswd"},
        )

    def test_all_args(self) -> None:
        # patch runner.MakerBase.postOptions() so that 'basedir'
        # argument will not be converted to absolute path
        self.patch(runner.MakerBase, "postOptions", mock.Mock())

        opts = self.parse(
            "--force",
            "--relocatable",
            "--no-logrotate",
            "--keepalive=4",
            "--umask=0o22",
            "--maxdelay=3",
            "--numcpus=4",
            "--log-size=2",
            "--log-count=1",
            "--allow-shutdown=file",
            *self.req_args,
        )
        self.assertOptions(
            opts,
            {
                "force": True,
                "relocatable": True,
                "no-logrotate": True,
                "umask": "0o22",
                "maxdelay": 3,
                "numcpus": "4",
                "log-size": 2,
                "log-count": "1",
                "allow-shutdown": "file",
                "basedir": "bdir",
                "host": "mstr",
                "port": 5678,
                "name": "name",
                "passwd": "pswd",
            },
        )

    def test_master_url(self) -> None:
        with self.assertRaisesRegex(usage.UsageError, "<master> is not a URL - do not use URL"):
            self.parse("a", "http://b.c", "d", "e")

    def test_inv_keepalive(self) -> None:
        with self.assertRaisesRegex(usage.UsageError, "keepalive parameter needs to be a number"):
            self.parse("--keepalive=X", *self.req_args)

    def test_inv_maxdelay(self) -> None:
        with self.assertRaisesRegex(usage.UsageError, "maxdelay parameter needs to be a number"):
            self.parse("--maxdelay=X", *self.req_args)

    def test_inv_log_size(self) -> None:
        with self.assertRaisesRegex(usage.UsageError, "log-size parameter needs to be a number"):
            self.parse("--log-size=X", *self.req_args)

    def test_inv_log_count(self) -> None:
        with self.assertRaisesRegex(
            usage.UsageError, "log-count parameter needs to be a number or None"
        ):
            self.parse("--log-count=X", *self.req_args)

    def test_inv_numcpus(self) -> None:
        with self.assertRaisesRegex(
            usage.UsageError, "numcpus parameter needs to be a number or None"
        ):
            self.parse("--numcpus=X", *self.req_args)

    def test_inv_umask(self) -> None:
        with self.assertRaisesRegex(
            usage.UsageError, "umask parameter needs to be a number or None"
        ):
            self.parse("--umask=X", *self.req_args)

    def test_inv_umask2(self) -> None:
        with self.assertRaisesRegex(
            usage.UsageError, "umask parameter needs to be a number or None"
        ):
            self.parse("--umask=022", *self.req_args)

    def test_inv_allow_shutdown(self) -> None:
        with self.assertRaisesRegex(
            usage.UsageError, "allow-shutdown needs to be one of 'signal' or 'file'"
        ):
            self.parse("--allow-shutdown=X", *self.req_args)

    def test_too_few_args(self) -> None:
        with self.assertRaisesRegex(usage.UsageError, "incorrect number of arguments"):
            self.parse("arg1", "arg2")

    def test_too_many_args(self) -> None:
        with self.assertRaisesRegex(usage.UsageError, "incorrect number of arguments"):
            self.parse("extra_arg", *self.req_args)

    def test_validateMasterArgument_no_port(self) -> None:
        """
        test calling CreateWorkerOptions.validateMasterArgument()
        on <master> argument without port specified.
        """
        opts = runner.CreateWorkerOptions()
        self.assertEqual(
            opts.validateMasterArgument("mstrhost"),
            ("mstrhost", 9989),
            "incorrect master host and/or port",
        )

    def test_validateMasterArgument_empty_master(self) -> None:
        """
        test calling CreateWorkerOptions.validateMasterArgument()
        on <master> without host part specified.
        """
        opts = runner.CreateWorkerOptions()
        with self.assertRaisesRegex(usage.UsageError, "invalid <master> argument ':1234'"):
            opts.validateMasterArgument(":1234")

    def test_validateMasterArgument_inv_port(self) -> None:
        """
        test calling CreateWorkerOptions.validateMasterArgument()
        on <master> without with unparsable port part
        """
        opts = runner.CreateWorkerOptions()
        with self.assertRaisesRegex(
            usage.UsageError, "invalid master port 'apple', needs to be a number"
        ):
            opts.validateMasterArgument("host:apple")

    def test_validateMasterArgument_ok(self) -> None:
        """
        test calling CreateWorkerOptions.validateMasterArgument()
        on <master> with host and port parts specified.
        """
        opts = runner.CreateWorkerOptions()
        self.assertEqual(
            opts.validateMasterArgument("mstrhost:4321"),
            ("mstrhost", 4321),
            "incorrect master host and/or port",
        )

    def test_validateMasterArgument_ipv4(self) -> None:
        """
        test calling CreateWorkerOptions.validateMasterArgument()
        on <master> with ipv4 host specified.
        """
        opts = runner.CreateWorkerOptions()
        self.assertEqual(
            opts.validateMasterArgument("192.168.0.0"),
            ("192.168.0.0", 9989),
            "incorrect master host and/or port",
        )

    def test_validateMasterArgument_ipv4_port(self) -> None:
        """
        test calling CreateWorkerOptions.validateMasterArgument()
        on <master> with ipv4 host and port parts specified.
        """
        opts = runner.CreateWorkerOptions()
        self.assertEqual(
            opts.validateMasterArgument("192.168.0.0:4321"),
            ("192.168.0.0", 4321),
            "incorrect master host and/or port",
        )

    def test_validateMasterArgument_ipv6(self) -> None:
        """
        test calling CreateWorkerOptions.validateMasterArgument()
        on <master> with ipv6 host specified.
        """
        opts = runner.CreateWorkerOptions()
        self.assertEqual(
            opts.validateMasterArgument("[2001:1:2:3:4::1]"),
            ("2001:1:2:3:4::1", 9989),
            "incorrect master host and/or port",
        )

    def test_validateMasterArgument_ipv6_port(self) -> None:
        """
        test calling CreateWorkerOptions.validateMasterArgument()
        on <master> with ipv6 host and port parts specified.
        """
        opts = runner.CreateWorkerOptions()
        self.assertEqual(
            opts.validateMasterArgument("[2001:1:2:3:4::1]:4321"),
            ("2001:1:2:3:4::1", 4321),
            "incorrect master host and/or port",
        )

    def test_validateMasterArgument_ipv6_no_bracket(self) -> None:
        """
        test calling CreateWorkerOptions.validateMasterArgument()
        on <master> with ipv6 without [] specified.
        """
        opts = runner.CreateWorkerOptions()
        with self.assertRaisesRegex(
            usage.UsageError,
            r"invalid <master> argument '2001:1:2:3:4::1:4321', "
            r"if it is an ipv6 address, it must be enclosed by \[\]",
        ):
            opts.validateMasterArgument("2001:1:2:3:4::1:4321")


class TestOptions(misc.StdoutAssertionsMixin, unittest.TestCase):
    """
    Test buildbot_worker.scripts.runner.Options class.
    """

    def setUp(self) -> None:
        self.setUpStdoutAssertions()

    def parse(self, *args: str) -> runner.Options:
        opts = runner.Options()
        opts.parseOptions(args)
        return opts

    def test_defaults(self) -> None:
        with self.assertRaisesRegex(usage.UsageError, "must specify a command"):
            self.parse()

    def test_version(self) -> None:
        exception = self.assertRaises(SystemExit, self.parse, '--version')
        self.assertEqual(exception.code, 0, "unexpected exit code")
        self.assertInStdout('worker version:')

    def test_verbose(self) -> None:
        self.patch(log, 'startLogging', mock.Mock())
        with self.assertRaises(usage.UsageError):
            self.parse("--verbose")
        cast("mock.Mock", log.startLogging).assert_called_once_with(sys.stderr)


# used by TestRun.test_run_good to patch in a callback
functionPlaceholder: Callable | None = None


class TestRun(misc.StdoutAssertionsMixin, unittest.TestCase):
    """
    Test buildbot_worker.scripts.runner.run()
    """

    def setUp(self) -> None:
        self.setUpStdoutAssertions()

    class TestSubCommand(usage.Options):
        subcommandFunction = __name__ + ".functionPlaceholder"
        optFlags = [["test-opt", None, None]]

    class TestOptions(usage.Options):
        """
        Option class that emulates usage error. The 'suboptions' flag
        enables emulation of usage error in a sub-option.
        """

        optFlags = [["suboptions", None, None]]

        def postOptions(self) -> NoReturn:
            if self["suboptions"]:
                self.subOptions = "SubOptionUsage"
            raise usage.UsageError("usage-error-message")

        def __str__(self) -> str:
            return "GeneralUsage"

    def test_run_good(self) -> None:
        """
        Test successful invocation of worker command.
        """

        self.patch(sys, "argv", ["command", 'test', '--test-opt'])

        # patch runner module to use our test subcommand class
        self.patch(runner.Options, 'subCommands', [['test', None, self.TestSubCommand, None]])

        # trace calls to subcommand function
        subcommand_func = mock.Mock(return_value=42)
        self.patch(sys.modules[__name__], "functionPlaceholder", subcommand_func)

        # check that subcommand function called with correct arguments
        # and that it's return value is used as exit code
        exception = self.assertRaises(SystemExit, runner.run)
        subcommand_func.assert_called_once_with({'test-opt': 1})
        self.assertEqual(exception.code, 42, "unexpected exit code")

    def test_run_bad_noargs(self) -> None:
        """
        Test handling of invalid command line arguments.
        """
        self.patch(sys, "argv", ["command"])

        # patch runner module to use test Options class
        self.patch(runner, "Options", self.TestOptions)

        exception = self.assertRaises(SystemExit, runner.run)
        self.assertEqual(exception.code, 1, "unexpected exit code")
        self.assertStdoutEqual(
            "command:  usage-error-message\n\nGeneralUsage\n",
            "unexpected error message on stdout",
        )

    def test_run_bad_suboption(self) -> None:
        """
        Test handling of invalid command line arguments in a suboption.
        """

        self.patch(sys, "argv", ["command", "--suboptions"])

        # patch runner module to use test Options class
        self.patch(runner, "Options", self.TestOptions)

        exception = self.assertRaises(SystemExit, runner.run)
        self.assertEqual(exception.code, 1, "unexpected exit code")

        # check that we get error message for a sub-option
        self.assertStdoutEqual(
            "command:  usage-error-message\n\nSubOptionUsage\n",
            "unexpected error message on stdout",
        )



"""
Protocol wrapper that will detect hung connections.

In particular, since PB expects the server to talk first and HTTP
expects the client to talk first, when a PB client talks to an HTTP
server, neither side will talk, leading to a hung connection. This
wrapper will disconnect in that case, and inform the caller.
"""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import cast

from twisted.internet.interfaces import IProtocol
from twisted.internet.interfaces import IProtocolFactory
from twisted.python.components import proxyForInterface

if TYPE_CHECKING:
    from typing import Callable

    from twisted.internet.interfaces import IAddress
    from twisted.internet.interfaces import IConnector
    from twisted.internet.interfaces import IDelayedCall
    from twisted.internet.interfaces import IReactorTime
    from twisted.internet.interfaces import ITransport
    from twisted.internet.protocol import ClientFactory
    from twisted.python.failure import Failure


def _noop() -> None:
    pass


class HangCheckProtocol(
    proxyForInterface(IProtocol, '_wrapped_protocol'),  # type: ignore[misc]
):
    """
    Wrap a protocol, so the underlying connection will disconnect if
    the other end doesn't send data within a given timeout.
    """

    transport: ITransport | None = None
    _hungConnectionTimer: IDelayedCall | None = None

    # hung connections wait for a relatively long time, since a busy master may
    # take a while to get back to us.
    _HUNG_CONNECTION_TIMEOUT = 120

    def __init__(
        self,
        wrapped_protocol: IProtocol,
        hung_callback: Callable[[], None] = _noop,
        reactor: IReactorTime | None = None,
    ) -> None:
        """
        :param IProtocol wrapped_protocol: The protocol to wrap.
        :param hung_callback: Called when the connection has hung.
        :type hung_callback: callable taking no arguments.
        :param IReactorTime reactor: The reactor to use to schedule
            the hang check.
        """
        if reactor is None:
            from twisted.internet import reactor as default_reactor

            reactor = cast("IReactorTime", default_reactor)

        self._wrapped_protocol = wrapped_protocol
        self._reactor = reactor
        self._hung_callback = hung_callback

    def makeConnection(self, transport: ITransport) -> None:
        # Note that we don't wrap the transport for the protocol,
        # because we only care about noticing data received, not
        # sent.
        self.transport = transport
        super().makeConnection(transport)
        self._startHungConnectionTimer()

    def dataReceived(self, data: bytes) -> None:
        self._stopHungConnectionTimer()
        super().dataReceived(data)

    def connectionLost(self, reason: Failure) -> None:
        self._stopHungConnectionTimer()
        super().connectionLost(reason)

    def _startHungConnectionTimer(self) -> None:
        """
        Start a timer to detect if the connection is hung.
        """

        def hungConnection() -> None:
            self._hung_callback()
            self._hungConnectionTimer = None
            if self.transport is not None:
                self.transport.loseConnection()

        self._hungConnectionTimer = self._reactor.callLater(
            self._HUNG_CONNECTION_TIMEOUT, hungConnection
        )

    def _stopHungConnectionTimer(self) -> None:
        """
        Cancel the hang check timer, since we have received data or
        been closed.
        """
        if self._hungConnectionTimer:
            self._hungConnectionTimer.cancel()
        self._hungConnectionTimer = None


class HangCheckFactory(
    proxyForInterface(IProtocolFactory, '_wrapped_factory'),  # type: ignore[misc]
):
    """
    Wrap a protocol factory, so the underlying connection will
    disconnect if the other end doesn't send data within a given
    timeout.
    """

    def __init__(
        self,
        wrapped_factory: ClientFactory,
        hung_callback: Callable[[], None],
    ) -> None:
        """
        :param IProtocolFactory wrapped_factory: The factory to wrap.
        :param hung_callback: Called when the connection has hung.
        :type hung_callback: callable taking no arguments.
        """
        self._wrapped_factory = wrapped_factory
        self._hung_callback = hung_callback

    def buildProtocol(self, addr: IAddress) -> HangCheckProtocol | None:
        protocol = self._wrapped_factory.buildProtocol(addr)
        if protocol is None:
            return None
        return HangCheckProtocol(protocol, hung_callback=self._hung_callback)

    # This is used as a ClientFactory, which doesn't have a specific interface, so forward the
    # additional methods.

    def startedConnecting(self, connector: IConnector) -> None:
        self._wrapped_factory.startedConnecting(connector)

    def clientConnectionFailed(self, connector: IConnector, reason: Failure) -> None:
        self._wrapped_factory.clientConnectionFailed(connector, reason)

    def clientConnectionLost(self, connector: IConnector, reason: Failure) -> None:
        self._wrapped_factory.clientConnectionLost(connector, reason)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import os
import time

import sqlalchemy as sa
from twisted.internet import defer
from twisted.internet import reactor
from twisted.trial import unittest

from buildbot.db import enginestrategy
from buildbot.db import pool
from buildbot.test.util import db
from buildbot.test.util.db import thd_clean_database
from buildbot.util import sautils


class Basic(unittest.TestCase):
    # basic tests, just using an in-memory SQL db and one thread

    @defer.inlineCallbacks
    def setUp(self):
        url = db.resolve_test_db_url(None, sqlite_memory=True)
        self.engine = enginestrategy.create_engine(url, basedir=os.getcwd())
        self.engine.should_retry = lambda _: False
        self.engine.optimal_thread_pool_size = 1
        self.pool = pool.DBThreadPool(self.engine, reactor=reactor)
        self.pool.start()
        yield self.pool.do(thd_clean_database)
        self.addCleanup(self.pool.stop)

    @defer.inlineCallbacks
    def test_do(self):
        def add(conn, addend1, addend2):
            rp = conn.execute(sa.text(f"SELECT {addend1} + {addend2}"))
            return rp.scalar()

        res = yield self.pool.do(add, 10, 11)

        self.assertEqual(res, 21)

    @defer.inlineCallbacks
    def expect_failure(self, d, expected_exceptions, expect_logged_error=False):
        exception = None
        try:
            yield d
        except Exception as e:
            exception = e
        errors = []
        for expected_exception in expected_exceptions:
            errors += self.flushLoggedErrors(expected_exception)
        if expect_logged_error:
            self.assertEqual(len(errors), 1)
        self.assertTrue(isinstance(exception, expected_exceptions))

    def test_do_error(self):
        def fail(conn):
            rp = conn.execute(sa.text("EAT COOKIES"))
            return rp.scalar()

        return self.expect_failure(
            self.pool.do(fail),
            (sa.exc.ProgrammingError, sa.exc.OperationalError),
            expect_logged_error=True,
        )

    def test_do_exception(self):
        def raise_something(conn):
            raise RuntimeError("oh noes")

        return self.expect_failure(
            self.pool.do(raise_something), (RuntimeError,), expect_logged_error=True
        )

    @defer.inlineCallbacks
    def test_do_with_engine(self):
        def add(engine, addend1, addend2):
            with engine.connect() as conn:
                rp = conn.execute(sa.text(f"SELECT {addend1} + {addend2}"))
                return rp.scalar()

        res = yield self.pool.do_with_engine(add, 10, 11)

        self.assertEqual(res, 21)

    def test_do_with_engine_exception(self):
        def fail(engine):
            with engine.connect() as conn:
                rp = conn.execute(sa.text("EAT COOKIES"))
            return rp.scalar()

        return self.expect_failure(
            self.pool.do_with_engine(fail), (sa.exc.ProgrammingError, sa.exc.OperationalError)
        )

    @defer.inlineCallbacks
    def test_persistence_across_invocations(self):
        # NOTE: this assumes that both methods are called with the same
        # connection; if they run in parallel threads then it is not valid to
        # assume that the database engine will have finalized the first
        # transaction (and thus created the table) by the time the second
        # transaction runs.  This is why we set optimal_thread_pool_size in
        # setUp.
        def create_table(engine):
            with engine.connect() as conn:
                conn.execute(sa.text("CREATE TABLE tmp ( a integer )"))
                conn.commit()

        yield self.pool.do_with_engine(create_table)

        def insert_into_table(engine):
            with engine.connect() as conn:
                conn.execute(sa.text("INSERT INTO tmp values ( 1 )"))
                conn.commit()

        yield self.pool.do_with_engine(insert_into_table)

    @defer.inlineCallbacks
    def test_ddl_and_queries(self):
        meta = sa.MetaData()
        native_tests = sautils.Table("native_tests", meta, sa.Column('name', sa.String(length=200)))

        # perform a DDL operation and immediately try to access that table;
        # this has caused problems in the past, so this is basically a
        # regression test.
        def ddl(conn):
            t = conn.begin()
            native_tests.create(bind=conn)
            t.commit()

        yield self.pool.do(ddl)

        def access(conn):
            conn.execute(native_tests.insert().values({'name': 'foo'}))

        yield self.pool.do_with_transaction(access)


class Stress(unittest.TestCase):
    def setUp(self):
        setup_engine = sa.create_engine('sqlite:///test.sqlite', future=True)
        setup_engine.execute("pragma journal_mode = wal")
        setup_engine.execute("CREATE TABLE test (a integer, b integer)")

        self.engine = sa.create_engine('sqlite:///test.sqlite', future=True)
        self.engine.optimal_thread_pool_size = 2
        self.pool = pool.DBThreadPool(self.engine, reactor=reactor)
        self.pool.start()
        self.addCleanup(self.pool.stop)

    def tearDown(self):
        os.unlink("test.sqlite")

    @defer.inlineCallbacks
    def test_inserts(self):
        def write(conn):
            trans = conn.begin()
            conn.execute("INSERT INTO test VALUES (1, 1)")
            time.sleep(31)
            trans.commit()

        d1 = self.pool.do(write)

        def write2(conn):
            trans = conn.begin()
            conn.execute("INSERT INTO test VALUES (1, 1)")
            trans.commit()

        d2 = defer.Deferred()
        d2.addCallback(lambda _: self.pool.do(write2))
        reactor.callLater(0.1, d2.callback, None)

        yield defer.DeferredList([d1, d2], consumeErrors=True)

    # don't run this test, since it takes 30s
    del test_inserts


class BasicWithDebug(Basic):
    # same thing, but with debug=True

    def setUp(self):
        pool.debug = True
        return super().setUp()

    def tearDown(self):
        pool.debug = False
        return super().tearDown()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import errno
import os
import string
import textwrap
from io import StringIO

from twisted.python import runtime
from twisted.python import usage
from twisted.trial import unittest

from buildbot.config import master as config_master
from buildbot.scripts import base
from buildbot.test.util import dirs
from buildbot.test.util import misc
from buildbot.test.util.decorators import skipUnlessPlatformIs


class TestIBD(dirs.DirsMixin, misc.StdoutAssertionsMixin, unittest.TestCase):
    def setUp(self):
        self.setUpDirs('test')
        self.stdout = StringIO()
        self.setUpStdoutAssertions()

    def test_isBuildmasterDir_no_dir(self):
        self.assertFalse(base.isBuildmasterDir(os.path.abspath('test/nosuch')))
        self.assertInStdout('error reading')
        self.assertInStdout('invalid buildmaster directory')

    def test_isBuildmasterDir_no_file(self):
        self.assertFalse(base.isBuildmasterDir(os.path.abspath('test')))
        self.assertInStdout('error reading')
        self.assertInStdout('invalid buildmaster directory')

    def test_isBuildmasterDir_no_Application(self):
        # Loading of pre-0.9.0 buildbot.tac file should fail.
        with open(os.path.join('test', 'buildbot.tac'), 'w', encoding='utf-8') as f:
            f.write("foo\nx = Application('buildslave')\nbar")
        self.assertFalse(base.isBuildmasterDir(os.path.abspath('test')))
        self.assertInStdout('unexpected content')
        self.assertInStdout('invalid buildmaster directory')

    def test_isBuildmasterDir_matches(self):
        with open(os.path.join('test', 'buildbot.tac'), 'w', encoding='utf-8') as f:
            f.write("foo\nx = Application('buildmaster')\nbar")
        self.assertTrue(base.isBuildmasterDir(os.path.abspath('test')))
        self.assertWasQuiet()


class TestTacFallback(dirs.DirsMixin, unittest.TestCase):
    """
    Tests for L{base.getConfigFileFromTac}.
    """

    def setUp(self):
        """
        Create a base directory.
        """
        self.basedir = os.path.abspath('basedir')
        return self.setUpDirs('basedir')

    def _createBuildbotTac(self, contents=None):
        """
        Create a C{buildbot.tac} that points to a given C{configfile}
        and create that file.

        @param configfile: Config file to point at and create.
        @type configfile: L{str}
        """
        if contents is None:
            contents = '#dummy'
        tacfile = os.path.join(self.basedir, "buildbot.tac")
        with open(tacfile, "w", encoding='utf-8') as f:
            f.write(contents)
        return tacfile

    def test_getConfigFileFromTac(self):
        """
        When L{getConfigFileFromTac} is passed a C{basedir}
        containing a C{buildbot.tac}, it reads the location
        of the config file from there.
        """
        self._createBuildbotTac("configfile='other.cfg'")
        foundConfigFile = base.getConfigFileFromTac(basedir=self.basedir)
        self.assertEqual(foundConfigFile, "other.cfg")

    def test_getConfigFileFromTac_fallback(self):
        """
        When L{getConfigFileFromTac} is passed a C{basedir}
        which doesn't contain a C{buildbot.tac},
        it returns C{master.cfg}
        """
        foundConfigFile = base.getConfigFileFromTac(basedir=self.basedir)
        self.assertEqual(foundConfigFile, 'master.cfg')

    def test_getConfigFileFromTac_tacWithoutConfigFile(self):
        """
        When L{getConfigFileFromTac} is passed a C{basedir}
        containing a C{buildbot.tac}, but C{buildbot.tac} doesn't
        define C{configfile}, L{getConfigFileFromTac} returns C{master.cfg}
        """
        self._createBuildbotTac()
        foundConfigFile = base.getConfigFileFromTac(basedir=self.basedir)
        self.assertEqual(foundConfigFile, 'master.cfg')

    def test_getConfigFileFromTac_usingFile(self):
        """
        When L{getConfigFileFromTac} is passed a C{basedir}
        containing a C{buildbot.tac} which references C{__file__},
        that reference points to C{buildbot.tac}.
        """
        self._createBuildbotTac(
            textwrap.dedent("""
            from twisted.python.util import sibpath
            configfile = sibpath(__file__, "relative.cfg")
            """)
        )
        foundConfigFile = base.getConfigFileFromTac(basedir=self.basedir)
        self.assertEqual(foundConfigFile, os.path.join(self.basedir, "relative.cfg"))


class TestSubcommandOptions(unittest.TestCase):
    def fakeOptionsFile(self, **kwargs):
        self.patch(base.SubcommandOptions, 'loadOptionsFile', lambda self: kwargs.copy())

    def parse(self, cls, *args):
        self.opts = cls()
        self.opts.parseOptions(args)
        return self.opts

    class Bare(base.SubcommandOptions):
        optFlags = [['foo', 'f', 'Foo!']]

    def test_bare_subclass(self):
        self.fakeOptionsFile()
        opts = self.parse(self.Bare, '-f')
        self.assertTrue(opts['foo'])

    class ParamsAndOptions(base.SubcommandOptions):
        optParameters = [['volume', 'v', '5', 'How Loud?']]
        buildbotOptions = [['volcfg', 'volume']]

    def test_buildbotOptions(self):
        self.fakeOptionsFile()
        opts = self.parse(self.ParamsAndOptions)
        self.assertEqual(opts['volume'], '5')

    def test_buildbotOptions_options(self):
        self.fakeOptionsFile(volcfg='3')
        opts = self.parse(self.ParamsAndOptions)
        self.assertEqual(opts['volume'], '3')

    def test_buildbotOptions_override(self):
        self.fakeOptionsFile(volcfg='3')
        opts = self.parse(self.ParamsAndOptions, '--volume', '7')
        self.assertEqual(opts['volume'], '7')

    class RequiredOptions(base.SubcommandOptions):
        optParameters = [['volume', 'v', None, 'How Loud?']]
        requiredOptions = ['volume']

    def test_requiredOptions(self):
        self.fakeOptionsFile()
        with self.assertRaises(usage.UsageError):
            self.parse(self.RequiredOptions)


class TestLoadOptionsFile(dirs.DirsMixin, misc.StdoutAssertionsMixin, unittest.TestCase):
    def setUp(self):
        self.setUpDirs('test', 'home')
        self.opts = base.SubcommandOptions()
        self.dir = os.path.abspath('test')
        self.home = os.path.abspath('home')
        self.setUpStdoutAssertions()

    def do_loadOptionsFile(self, _here, exp):
        # only patch these os.path functions briefly, to
        # avoid breaking other parts of the test system
        patches = []

        if runtime.platformType == 'win32':
            from win32com.shell import shell

            patches.append(self.patch(shell, 'SHGetFolderPath', lambda *args: self.home))
        else:

            def expanduser(p):
                return p.replace('~/', self.home + '/')

            patches.append(self.patch(os.path, 'expanduser', expanduser))

        old_dirname = os.path.dirname

        def dirname(p):
            # bottom out at self.dir, rather than /
            if p == self.dir:
                return p
            return old_dirname(p)

        patches.append(self.patch(os.path, 'dirname', dirname))

        try:
            self.assertEqual(self.opts.loadOptionsFile(_here=_here), exp)
        finally:
            for p in patches:
                p.restore()

    def writeOptionsFile(self, dir, content, bbdir='.buildbot'):
        os.makedirs(os.path.join(dir, bbdir))
        with open(os.path.join(dir, bbdir, 'options'), 'w', encoding='utf-8') as f:
            f.write(content)

    def test_loadOptionsFile_subdirs_not_found(self):
        subdir = os.path.join(self.dir, 'a', 'b')
        os.makedirs(subdir)
        self.do_loadOptionsFile(_here=subdir, exp={})

    def test_loadOptionsFile_subdirs_at_root(self):
        subdir = os.path.join(self.dir, 'a', 'b')
        os.makedirs(subdir)
        self.writeOptionsFile(self.dir, 'abc="def"')
        self.writeOptionsFile(self.home, 'abc=123')  # not seen
        self.do_loadOptionsFile(_here=subdir, exp={'abc': 'def'})

    def test_loadOptionsFile_subdirs_at_tip(self):
        subdir = os.path.join(self.dir, 'a', 'b')
        os.makedirs(subdir)
        self.writeOptionsFile(os.path.join(self.dir, 'a', 'b'), 'abc="def"')
        self.writeOptionsFile(self.dir, 'abc=123')  # not seen
        self.do_loadOptionsFile(_here=subdir, exp={'abc': 'def'})

    def test_loadOptionsFile_subdirs_at_homedir(self):
        subdir = os.path.join(self.dir, 'a', 'b')
        os.makedirs(subdir)
        # on windows, the subdir of the home (well, appdata) dir
        # is 'buildbot', not '.buildbot'
        self.writeOptionsFile(
            self.home, 'abc=123', 'buildbot' if runtime.platformType == 'win32' else '.buildbot'
        )
        self.do_loadOptionsFile(_here=subdir, exp={'abc': 123})

    def test_loadOptionsFile_syntax_error(self):
        self.writeOptionsFile(self.dir, 'abc=abc')
        with self.assertRaises(NameError):
            self.do_loadOptionsFile(_here=self.dir, exp={})
        self.assertInStdout('error while reading')

    def test_loadOptionsFile_toomany(self):
        subdir = os.path.join(self.dir, *tuple(string.ascii_lowercase))
        os.makedirs(subdir)
        self.do_loadOptionsFile(_here=subdir, exp={})
        self.assertInStdout('infinite glories')

    # NOTE: testing the ownership check requires patching os.stat, which causes
    # other problems since it is so heavily used.


def mkconfig(**kwargs):
    config = {"quiet": False, "replace": False, "basedir": 'test'}
    config.update(kwargs)
    return config


class TestLoadConfig(dirs.DirsMixin, misc.StdoutAssertionsMixin, unittest.TestCase):
    def setUp(self):
        self.setUpDirs('test')
        self.setUpStdoutAssertions()

    def activeBasedir(self, extra_lines=()):
        with open(os.path.join('test', 'buildbot.tac'), "w", encoding='utf-8') as f:
            f.write("from twisted.application import service\n")
            f.write("service.Application('buildmaster')\n")
            f.write("\n".join(extra_lines))

    def test_checkBasedir(self):
        self.activeBasedir()
        rv = base.checkBasedir(mkconfig())
        self.assertTrue(rv)
        self.assertInStdout('checking basedir')

    def test_checkBasedir_quiet(self):
        self.activeBasedir()
        rv = base.checkBasedir(mkconfig(quiet=True))
        self.assertTrue(rv)
        self.assertWasQuiet()

    def test_checkBasedir_no_dir(self):
        rv = base.checkBasedir(mkconfig(basedir='doesntexist'))
        self.assertFalse(rv)
        self.assertInStdout('invalid buildmaster directory')

    @skipUnlessPlatformIs('posix')
    def test_checkBasedir_active_pidfile(self):
        """
        active PID file is giving error.
        """
        self.activeBasedir()
        # write our own pid in the file
        with open(os.path.join('test', 'twistd.pid'), 'w', encoding='utf-8') as f:
            f.write(str(os.getpid()))
        rv = base.checkBasedir(mkconfig())
        self.assertFalse(rv)
        self.assertInStdout('still running')

    @skipUnlessPlatformIs('posix')
    def test_checkBasedir_bad_pidfile(self):
        """
        corrupted PID file is giving error.
        """
        self.activeBasedir()
        with open(os.path.join('test', 'twistd.pid'), 'w', encoding='utf-8') as f:
            f.write("xxx")
        rv = base.checkBasedir(mkconfig())
        self.assertFalse(rv)
        self.assertInStdout('twistd.pid contains non-numeric value')

    @skipUnlessPlatformIs('posix')
    def test_checkBasedir_stale_pidfile(self):
        """
        Stale PID file is removed without causing a system exit.
        """
        self.activeBasedir()
        pidfile = os.path.join('test', 'twistd.pid')
        with open(pidfile, 'w', encoding='utf-8') as f:
            f.write(str(os.getpid() + 1))

        def kill(pid, sig):
            raise OSError(errno.ESRCH, "fake")

        self.patch(os, "kill", kill)
        rv = base.checkBasedir(mkconfig())
        self.assertTrue(rv)
        self.assertInStdout('Removing stale pidfile test')
        self.assertFalse(os.path.exists(pidfile))

    @skipUnlessPlatformIs('posix')
    def test_checkBasedir_pidfile_kill_error(self):
        """
        if ping-killing the PID file does not work, we should error out.
        """
        self.activeBasedir()
        # write our own pid in the file
        pidfile = os.path.join('test', 'twistd.pid')
        with open(pidfile, 'w', encoding='utf-8') as f:
            f.write(str(os.getpid() + 1))

        def kill(pid, sig):
            raise OSError(errno.EPERM, "fake")

        self.patch(os, "kill", kill)
        rv = base.checkBasedir(mkconfig())
        self.assertFalse(rv)
        self.assertInStdout('Can\'t check status of PID')
        self.assertTrue(os.path.exists(pidfile))

    def test_checkBasedir_invalid_rotateLength(self):
        self.activeBasedir(extra_lines=['rotateLength="32"'])
        rv = base.checkBasedir(mkconfig())
        self.assertFalse(rv)
        self.assertInStdout('ERROR')
        self.assertInStdout('rotateLength')

    def test_checkBasedir_invalid_maxRotatedFiles(self):
        self.activeBasedir(extra_lines=['maxRotatedFiles="64"'])
        rv = base.checkBasedir(mkconfig())
        self.assertFalse(rv)
        self.assertInStdout('ERROR')
        self.assertInStdout('maxRotatedFiles')

    def test_loadConfig(self):
        @classmethod
        def loadConfig(cls):
            return config_master.MasterConfig()

        self.patch(config_master.FileLoader, 'loadConfig', loadConfig)
        cfg = base.loadConfig(mkconfig())
        self.assertIsInstance(cfg, config_master.MasterConfig)
        self.assertInStdout('checking')

    def test_loadConfig_ConfigErrors(self):
        @classmethod
        def loadConfig(cls):
            raise config_master.ConfigErrors(['oh noes'])

        self.patch(config_master.FileLoader, 'loadConfig', loadConfig)
        cfg = base.loadConfig(mkconfig())
        self.assertIdentical(cfg, None)
        self.assertInStdout('oh noes')

    def test_loadConfig_exception(self):
        @classmethod
        def loadConfig(cls):
            raise RuntimeError()

        self.patch(config_master.FileLoader, 'loadConfig', loadConfig)
        cfg = base.loadConfig(mkconfig())
        self.assertIdentical(cfg, None)
        self.assertInStdout('RuntimeError')



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer
from twisted.trial import unittest

from buildbot.data import forceschedulers
from buildbot.schedulers.forcesched import ForceScheduler
from buildbot.test import fakedb
from buildbot.test.util import endpoint

expected_default = {
    'all_fields': [
        {
            'columns': 1,
            'autopopulate': None,
            'default': '',
            'fields': [
                {
                    'default': '',
                    'autopopulate': None,
                    'fullName': 'username',
                    'hide': False,
                    'label': 'Your name:',
                    'maxsize': None,
                    'multiple': False,
                    'name': 'username',
                    'need_email': True,
                    'regex': None,
                    'required': False,
                    'size': 30,
                    'tablabel': 'Your name:',
                    'type': 'username',
                    'tooltip': '',
                },
                {
                    'default': 'force build',
                    'autopopulate': None,
                    'fullName': 'reason',
                    'hide': False,
                    'label': 'reason',
                    'maxsize': None,
                    'multiple': False,
                    'name': 'reason',
                    'regex': None,
                    'required': False,
                    'size': 20,
                    'tablabel': 'reason',
                    'type': 'text',
                    'tooltip': '',
                },
                {
                    'default': 0,
                    'autopopulate': None,
                    'fullName': 'priority',
                    'hide': False,
                    'label': 'priority',
                    'maxsize': None,
                    'multiple': False,
                    'name': 'priority',
                    'regex': None,
                    'required': False,
                    'size': 10,
                    'tablabel': 'priority',
                    'type': 'int',
                    'tooltip': '',
                },
            ],
            'fullName': None,
            'hide': False,
            'label': '',
            'layout': 'vertical',
            'maxsize': None,
            'multiple': False,
            'name': '',
            'regex': None,
            'required': False,
            'tablabel': '',
            'type': 'nested',
            'tooltip': '',
        },
        {
            'columns': 2,
            'default': '',
            'fields': [
                {
                    'default': '',
                    'autopopulate': None,
                    'fullName': 'branch',
                    'hide': False,
                    'label': 'Branch:',
                    'multiple': False,
                    'maxsize': None,
                    'name': 'branch',
                    'regex': None,
                    'required': False,
                    'size': 10,
                    'tablabel': 'Branch:',
                    'type': 'text',
                    'tooltip': '',
                },
                {
                    'default': '',
                    'autopopulate': None,
                    'fullName': 'project',
                    'hide': False,
                    'label': 'Project:',
                    'maxsize': None,
                    'multiple': False,
                    'name': 'project',
                    'regex': None,
                    'required': False,
                    'size': 10,
                    'tablabel': 'Project:',
                    'type': 'text',
                    'tooltip': '',
                },
                {
                    'default': '',
                    'autopopulate': None,
                    'fullName': 'repository',
                    'hide': False,
                    'label': 'Repository:',
                    'maxsize': None,
                    'multiple': False,
                    'name': 'repository',
                    'regex': None,
                    'required': False,
                    'size': 10,
                    'tablabel': 'Repository:',
                    'type': 'text',
                    'tooltip': '',
                },
                {
                    'default': '',
                    'autopopulate': None,
                    'fullName': 'revision',
                    'hide': False,
                    'label': 'Revision:',
                    'maxsize': None,
                    'multiple': False,
                    'name': 'revision',
                    'regex': None,
                    'required': False,
                    'size': 10,
                    'tablabel': 'Revision:',
                    'type': 'text',
                    'tooltip': '',
                },
            ],
            'autopopulate': None,
            'fullName': None,
            'hide': False,
            'label': '',
            'layout': 'vertical',
            'maxsize': None,
            'multiple': False,
            'name': '',
            'regex': None,
            'required': False,
            'tablabel': '',
            'type': 'nested',
            'tooltip': '',
        },
    ],
    'builder_names': ['builder'],
    'button_name': 'defaultforce',
    'label': 'defaultforce',
    'name': 'defaultforce',
    'enabled': True,
}


class ForceschedulerEndpoint(endpoint.EndpointMixin, unittest.TestCase):
    endpointClass = forceschedulers.ForceSchedulerEndpoint
    resourceTypeClass = forceschedulers.ForceScheduler
    maxDiff = None

    @defer.inlineCallbacks
    def setUp(self):
        yield self.setUpEndpoint()
        scheds = [ForceScheduler(name="defaultforce", builderNames=["builder"])]
        self.master.allSchedulers = lambda: scheds
        for sched in scheds:
            yield sched.setServiceParent(self.master)
        yield self.master.db.insert_test_data([
            fakedb.Master(id=fakedb.FakeDBConnector.MASTER_ID),
        ])
        yield self.master.startService()

    @defer.inlineCallbacks
    def test_get_existing(self):
        res = yield self.callGet(('forceschedulers', "defaultforce"))
        self.validateData(res)
        self.assertEqual(res, expected_default)

    @defer.inlineCallbacks
    def test_get_missing(self):
        res = yield self.callGet(('forceschedulers', 'foo'))
        self.assertEqual(res, None)


class ForceSchedulersEndpoint(endpoint.EndpointMixin, unittest.TestCase):
    endpointClass = forceschedulers.ForceSchedulersEndpoint
    resourceTypeClass = forceschedulers.ForceScheduler
    maxDiff = None

    @defer.inlineCallbacks
    def setUp(self):
        yield self.setUpEndpoint()
        scheds = [ForceScheduler(name="defaultforce", builderNames=["builder"])]
        self.master.allSchedulers = lambda: scheds
        for sched in scheds:
            yield sched.setServiceParent(self.master)
        yield self.master.db.insert_test_data([
            fakedb.Master(id=fakedb.FakeDBConnector.MASTER_ID),
        ])
        yield self.master.startService()

    @defer.inlineCallbacks
    def test_get_existing(self):
        res = yield self.callGet(('forceschedulers',))
        self.assertEqual(res, [expected_default])



#!/usr/bin/env python
#
# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import buildbot_pkg
from setuptools import setup

setup(
    name='buildbot-pkg',
    version=buildbot_pkg.getVersion("."),
    description='Buildbot packaging tools',
    author='Pierre Tardy',
    author_email='tardyp@gmail.com',
    url='http://buildbot.net/',
    py_modules=['buildbot_pkg'],
    install_requires=[
        "setuptools >= 21.2.1",
    ],
    classifiers=['License :: OSI Approved :: GNU General Public License v2 (GPLv2)'],
)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

from twisted.internet import defer

from buildbot.data import base
from buildbot.data import types
from buildbot.data.resultspec import ResultSpec
from buildbot.util.twisted import async_to_deferred

if TYPE_CHECKING:
    from buildbot.db.builds import BuildModel
    from buildbot.util.twisted import InlineCallbacksType


def _db2data(model: BuildModel):
    return {
        'buildid': model.id,
        'number': model.number,
        'builderid': model.builderid,
        'buildrequestid': model.buildrequestid,
        'workerid': model.workerid,
        'masterid': model.masterid,
        'started_at': model.started_at,
        'complete_at': model.complete_at,
        "locks_duration_s": model.locks_duration_s,
        'complete': model.complete_at is not None,
        'state_string': model.state_string,
        'results': model.results,
        'properties': {},
    }


builds_field_map = {
    'buildid': 'builds.id',
    'number': 'builds.number',
    'builderid': 'builds.builderid',
    'buildrequestid': 'builds.buildrequestid',
    'workerid': 'builds.workerid',
    'masterid': 'builds.masterid',
    'started_at': 'builds.started_at',
    'complete_at': 'builds.complete_at',
    "locks_duration_s": "builds.locks_duration_s",
    'state_string': 'builds.state_string',
    'results': 'builds.results',
}


def _generate_filtered_properties(props, filters):
    """
    This method returns Build's properties according to property filters.

    .. seealso::

       `Official Documentation <https://docs.buildbot.net/latest/developer/rtype-build.html>`_

    :param props: The Build's properties as a dict (from db)
    :param filters: Desired properties keys as a list (from API URI)

    """
    # by default none properties are returned
    if props and filters:
        return props if '*' in filters else dict(((k, v) for k, v in props.items() if k in filters))
    return None


class BuildEndpoint(base.BuildNestingMixin, base.Endpoint):
    kind = base.EndpointKind.SINGLE
    pathPatterns = [
        "/builds/n:buildid",
        "/builders/n:builderid/builds/n:build_number",
        "/builders/s:buildername/builds/n:build_number",
    ]

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        if 'buildid' in kwargs:
            dbdict = yield self.master.db.builds.getBuild(kwargs['buildid'])
        else:
            bldr = yield self.getBuilderId(kwargs)
            if bldr is None:
                return None
            num = kwargs['build_number']
            dbdict = yield self.master.db.builds.getBuildByNumber(bldr, num)

        data = _db2data(dbdict) if dbdict else None
        # In some cases, data could be None
        if data:
            filters = resultSpec.popProperties() if hasattr(resultSpec, 'popProperties') else []
            # Avoid to request DB for Build's properties if not specified
            if filters:
                try:
                    props = yield self.master.db.builds.getBuildProperties(data['buildid'])
                except (KeyError, TypeError):
                    props = {}
                filtered_properties = _generate_filtered_properties(props, filters)
                if filtered_properties:
                    data['properties'] = filtered_properties
        return data

    @defer.inlineCallbacks
    def actionStop(self, args, kwargs):
        buildid = kwargs.get('buildid')
        if buildid is None:
            bldr = kwargs['builderid']
            num = kwargs['build_number']
            dbdict = yield self.master.db.builds.getBuildByNumber(bldr, num)
            buildid = dbdict.id
        self.master.mq.produce(
            ("control", "builds", str(buildid), 'stop'),
            {"reason": kwargs.get('reason', args.get('reason', 'no reason'))},
        )

    @defer.inlineCallbacks
    def actionRebuild(self, args, kwargs):
        # we use the self.get and not self.data.get to be able to support all
        # the pathPatterns of this endpoint
        build = yield self.get(ResultSpec(), kwargs)
        buildrequest = yield self.master.data.get(('buildrequests', build['buildrequestid']))
        res = yield self.master.data.updates.rebuildBuildrequest(buildrequest)
        return res


class BuildTriggeredBuildsEndpoint(base.Endpoint):
    kind = base.EndpointKind.COLLECTION
    pathPatterns = [
        "/builds/n:buildid/triggered_builds",
    ]

    @async_to_deferred
    async def get(self, result_spec: base.ResultSpec, kwargs: Any) -> list[dict[str, Any]]:
        builds = await self.master.db.builds.get_triggered_builds(kwargs['buildid'])
        return [_db2data(b) for b in builds]


class BuildsEndpoint(base.BuildNestingMixin, base.Endpoint):
    kind = base.EndpointKind.COLLECTION
    pathPatterns = [
        "/builds",
        "/builders/n:builderid/builds",
        "/builders/s:buildername/builds",
        "/buildrequests/n:buildrequestid/builds",
        "/changes/n:changeid/builds",
        "/workers/n:workerid/builds",
    ]
    rootLinkName = 'builds'

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        changeid = kwargs.get('changeid')
        if changeid is not None:
            builds = yield self.master.db.builds.getBuildsForChange(changeid)
        else:
            # following returns None if no filter
            # true or false, if there is a complete filter
            builderid = None
            if 'builderid' in kwargs or 'buildername' in kwargs:
                builderid = yield self.getBuilderId(kwargs)
                if builderid is None:
                    return []
            complete = resultSpec.popBooleanFilter("complete")
            buildrequestid = resultSpec.popIntegerFilter("buildrequestid")
            resultSpec.fieldMapping = builds_field_map
            builds = yield self.master.db.builds.getBuilds(
                builderid=builderid,
                buildrequestid=kwargs.get('buildrequestid', buildrequestid),
                workerid=kwargs.get('workerid'),
                complete=complete,
                resultSpec=resultSpec,
            )

        # returns properties' list
        filters = resultSpec.popProperties()

        buildscol = []
        for b in builds:
            data = _db2data(b)
            # Avoid to request DB for Build's properties if not specified
            if filters:
                props = yield self.master.db.builds.getBuildProperties(data["buildid"])
                filtered_properties = _generate_filtered_properties(props, filters)
                if filtered_properties:
                    data["properties"] = filtered_properties

            buildscol.append(data)
        return buildscol


class Build(base.ResourceType):
    name = "build"
    plural = "builds"
    endpoints = [BuildEndpoint, BuildsEndpoint, BuildTriggeredBuildsEndpoint]
    eventPathPatterns = [
        "/builders/:builderid/builds/:number",
        "/builds/:buildid",
        "/workers/:workerid/builds/:buildid",
    ]

    class EntityType(types.Entity):
        buildid = types.Integer()
        number = types.Integer()
        builderid = types.Integer()
        buildrequestid = types.Integer()
        workerid = types.Integer()
        masterid = types.Integer()
        started_at = types.DateTime()
        complete = types.Boolean()
        complete_at = types.NoneOk(types.DateTime())
        locks_duration_s = types.Integer()
        results = types.NoneOk(types.Integer())
        state_string = types.String()
        properties = types.NoneOk(types.SourcedProperties())

    entityType = EntityType(name)

    @defer.inlineCallbacks
    def generateEvent(self, _id, event):
        # get the build and munge the result for the notification
        build = yield self.master.data.get(('builds', str(_id)))
        self.produceEvent(build, event)

    @base.updateMethod
    @defer.inlineCallbacks
    def addBuild(
        self, builderid: int, buildrequestid: int, workerid: int
    ) -> InlineCallbacksType[tuple[int, int]]:
        res = yield self.master.db.builds.addBuild(
            builderid=builderid,
            buildrequestid=buildrequestid,
            workerid=workerid,
            masterid=self.master.masterid,
            state_string='created',
        )
        return res

    @base.updateMethod
    def generateNewBuildEvent(self, buildid: int) -> defer.Deferred:
        return self.generateEvent(buildid, "new")

    @base.updateMethod
    @defer.inlineCallbacks
    def setBuildStateString(self, buildid: int, state_string: str) -> InlineCallbacksType[None]:
        res = yield self.master.db.builds.setBuildStateString(
            buildid=buildid, state_string=state_string
        )
        yield self.generateEvent(buildid, "update")
        return res

    @base.updateMethod
    @defer.inlineCallbacks
    def add_build_locks_duration(self, buildid: int, duration_s: int) -> InlineCallbacksType[None]:
        yield self.master.db.builds.add_build_locks_duration(buildid=buildid, duration_s=duration_s)
        yield self.generateEvent(buildid, "update")

    @base.updateMethod
    @defer.inlineCallbacks
    def finishBuild(self, buildid: int, results: int) -> InlineCallbacksType[None]:
        res = yield self.master.db.builds.finishBuild(buildid=buildid, results=results)
        yield self.generateEvent(buildid, "finished")
        return res



#!/usr/bin/env python

# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Portions Copyright Buildbot Team Members
# Portions Copyright 2013 OpenGamma Inc. and the OpenGamma group of companies


import getpass
import optparse
import os
import textwrap
import urllib

import httplib

# Find a working json module.  Code is from
# Paul Wise <pabs@debian.org>:
#   http://lists.debian.org/debian-python/2010/02/msg00016.html
try:
    import json  # python 2.6

    assert json  # silence pyflakes
except ImportError:
    import simplejson as json  # python 2.4 to 2.5
try:
    _tmp = json.loads
except AttributeError:
    import sys
    import warnings

    warnings.warn(
        "Use simplejson, not the old json module.", category=DeprecationWarning, stacklevel=1
    )
    sys.modules.pop('json')  # get rid of the bad json module
    import simplejson as json

# Make a dictionary with options from command line


def buildURL(options):
    urlDict = {}
    if options.author:
        author = options.author
    else:
        author = getpass.getuser()

    urlDict['author'] = author

    if options.files:
        urlDict['files'] = json.dumps(options.files)

    if options.comments:
        urlDict['comments'] = options.comments
    else:
        # A comment is required by the buildbot DB
        urlDict['comments'] = 'post_build_request submission'

    if options.revision:
        urlDict['revision'] = options.revision

    if options.when:
        urlDict['when'] = options.when

    if options.branch:
        urlDict['branch'] = options.branch

    if options.category:
        urlDict['category'] = options.category

    if options.revlink:
        urlDict['revlink'] = options.revlink

    if options.properties:
        urlDict['properties'] = json.dumps(options.properties)

    if options.repository:
        urlDict['repository'] = options.repository

    if options.project:
        urlDict['project'] = options.project

    return urlDict


def propertyCB(option, opt, value, parser):
    pdict = eval(value)
    for key in pdict.keys():
        parser.values.properties[key] = pdict[key]


__version__ = '0.1'

description = ""

usage = """%prog [options]

This script is used to submit a change to the buildbot master using the
/change_hook web interface. Options are url encoded and submitted
using a HTTP POST. The repository and project must be specified.

This can be used to force a build. For example, create a scheduler that
listens for changes on a category 'release':

releaseFilt    = ChangeFilter(category="release")
s=Scheduler(name="Release", change_filter=releaseFilt,
            treeStableTimer=10,
            builderNames=["UB10.4 x86_64 Release"]))
c['schedulers'].append(s)

Then run this script with the options:

--repository <REPOSTORY> --project <PROJECT> --category release
"""

parser = optparse.OptionParser(
    description=description, usage=usage, add_help_option=True, version=__version__
)

parser.add_option(
    "-w",
    "--who",
    dest='author',
    metavar="AUTHOR",
    help=textwrap.dedent("""\
            Who is submitting this request.
            This becomes the Change.author attribute.
            This defaults to the name of the user running this script
            """),
)
parser.add_option(
    "-f",
    "--file",
    dest='files',
    action="append",
    metavar="FILE",
    help=textwrap.dedent("""\
            Add a file to the change request.
            This is added to the Change.files attribute.
            NOTE: Setting the file URL is not supported
            """),
)
parser.add_option(
    "-c",
    "--comments",
    dest='comments',
    metavar="COMMENTS",
    help=textwrap.dedent("""\
            Comments for the change. This becomes the Change.comments attribute
            """),
)
parser.add_option(
    "-R",
    "--revision",
    dest='revision',
    metavar="REVISION",
    help=textwrap.dedent("""\
            This is the revision of the change.
            This becomes the Change.revision attribute.
            """),
)
parser.add_option(
    "-W",
    "--when",
    dest='when',
    metavar="WHEN",
    help=textwrap.dedent("""\
            This this the date of the change.
            This becomes the Change.when attribute.
            """),
)
parser.add_option(
    "-b",
    "--branch",
    dest='branch',
    metavar="BRANCH",
    help=textwrap.dedent("""\
            This this the branch of the change.
            This becomes the Change.branch attribute.
            """),
)
parser.add_option(
    "-C",
    "--category",
    dest='category',
    metavar="CAT",
    help=textwrap.dedent("""\
            Category for change. This becomes the Change.category attribute, which
            can be used within the buildmaster to filter changes.
            """),
)
parser.add_option(
    "--revlink",
    dest='revlink',
    metavar="REVLINK",
    help=textwrap.dedent("""\
            This this the revlink of the change.
            This becomes the Change.revlink.
            """),
)
parser.add_option(
    "-p",
    "--property",
    dest='properties',
    action="callback",
    callback=propertyCB,
    type="string",
    metavar="PROP",
    help=textwrap.dedent("""\
            This adds a single property. This can be specified multiple times.
            The argument is a string representing python dictionary. For example,
            {'foo' : [ 'bar', 'baz' ]}
            This becomes the Change.properties attribute.
            """),
)
parser.add_option(
    "-r",
    "--repository",
    dest='repository',
    metavar="PATH",
    help=textwrap.dedent("""\
            Repository for use by buildbot workers to checkout code.
            This becomes the Change.repository attribute.
            Exmaple: :ext:myhost:/cvsroot
            """),
)
parser.add_option(
    "-P",
    "--project",
    dest='project',
    metavar="PROJ",
    help=textwrap.dedent("""\
            The project for the source. Often set to the CVS module being modified. This becomes
            the Change.project attribute.
            """),
)
parser.add_option(
    "-v",
    "--verbose",
    dest='verbosity',
    action="count",
    help=textwrap.dedent("""\
            Print more detail. Shows the response status and reason received from the master. If
            specified twice, it also shows the raw response.
            """),
)
parser.add_option(
    "-H",
    "--host",
    dest='host',
    metavar="HOST",
    default='localhost:8010',
    help=textwrap.dedent("""\
            Host and optional port of buildbot. For example, bbhost:8010
            Defaults to %default
            """),
)
parser.add_option(
    "-u",
    "--urlpath",
    dest='urlpath',
    metavar="URLPATH",
    default='/change_hook/base',
    help=textwrap.dedent("""\
            Path portion of URL. Defaults to %default
            """),
)
parser.add_option(
    "-t",
    "--testing",
    action="store_true",
    dest="amTesting",
    default=False,
    help=textwrap.dedent("""\
            Just print values and exit.
            """),
)
parser.set_defaults(properties={})

(options, args) = parser.parse_args()

if options.repository is None:
    print("repository must be specified")
    parser.print_usage()
    os._exit(2)

if options.project is None:
    print("project must be specified")
    parser.print_usage()
    os._exit(2)

urlDict = buildURL(options)

params = urllib.urlencode(urlDict)
headers = {"Content-type": "application/x-www-form-urlencoded", "Accept": "text/plain"}
if options.amTesting:
    print(f"params: {params}")
    print(f"host: {options.host}")
    print(f"urlpath: {options.urlpath}")
else:
    conn = httplib.HTTPConnection(options.host)
    conn.request("POST", options.urlpath, params, headers)
    response = conn.getresponse()
    data = response.read()
    exitCode = 0
    if response.status != 202:
        exitCode = 1
    if options.verbosity >= 1:
        print(response.status, response.reason)
        if options.verbosity >= 2:
            print(f"Raw response: {data}")
    conn.close()
    os._exit(exitCode)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import Any
from typing import cast

from parameterized import parameterized
from twisted.trial import unittest

from buildbot.data import codebases
from buildbot.data import resultspec
from buildbot.master import BuildMaster
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import endpoint
from buildbot.test.util import interfaces
from buildbot.util.twisted import async_to_deferred


class CodebaseEndpoint(endpoint.EndpointMixin, unittest.TestCase):
    endpointClass = codebases.CodebaseEndpoint
    resourceTypeClass = codebases.Codebase

    @async_to_deferred
    async def setUp(self) -> None:  # type: ignore[override]
        await self.setUpEndpoint()
        await self.master.db.insert_test_data([
            fakedb.Project(id=7, name='fake_project7'),
            fakedb.Codebase(id=13, projectid=7, name='codebase1', slug='slug_codebase1'),
        ])

    @async_to_deferred
    async def test_get_existing_id(self) -> None:
        codebase = await self.callGet(('codebases', 13))

        self.validateData(codebase)
        self.assertEqual(
            codebase,
            {'codebaseid': 13, 'name': 'codebase1', 'projectid': 7, 'slug': 'slug_codebase1'},
        )

    @async_to_deferred
    async def test_get_missing(self) -> None:
        codebase = await self.callGet(('codebases', 100))
        self.assertIsNone(codebase)


class CodebasesEndpoint(endpoint.EndpointMixin, unittest.TestCase):
    endpointClass = codebases.CodebasesEndpoint
    resourceTypeClass = codebases.Codebase

    @async_to_deferred
    async def setUp(self) -> None:  # type: ignore[override]
        await self.setUpEndpoint()
        await self.master.db.insert_test_data([
            fakedb.Project(id=7, name='fake_project7'),
            fakedb.Project(id=8, name='fake_project8'),
            fakedb.Codebase(id=13, projectid=7, name='codebase1', slug='slug_codebase1'),
            fakedb.Codebase(id=14, projectid=7, name='codebase2'),
            fakedb.Codebase(id=15, projectid=8, name='codebase3'),
        ])

    @parameterized.expand([
        ('no_filter', None, [13, 14, 15]),
        ('existing', 7, [13, 14]),
        ('not_existing', 9, []),
    ])
    @async_to_deferred
    async def test_get(
        self, name: str, projectid_filter: int | None, expected_codebaseids: list[str]
    ) -> None:
        result_spec = None
        if projectid_filter is not None:
            result_spec = resultspec.OptimisedResultSpec(
                filters=[resultspec.Filter('projectid', 'eq', [projectid_filter])]
            )

        codebases = await self.callGet(('codebases',), resultSpec=result_spec)

        for b in codebases:
            self.validateData(b)

        self.assertEqual(sorted([b['codebaseid'] for b in codebases]), expected_codebaseids)


class Codebase(interfaces.InterfaceTests, TestReactorMixin, unittest.TestCase):
    @async_to_deferred
    async def setUp(self) -> None:  # type: ignore[override]
        self.setup_test_reactor()
        self.master = await fakemaster.make_master(self, wantMq=True, wantDb=True, wantData=True)
        self.rtype = codebases.Codebase(cast(BuildMaster, self.master))
        await self.master.db.insert_test_data([
            fakedb.Project(id=7, name='fake_project7'),
            fakedb.Project(id=8, name='fake_project8'),
            fakedb.Codebase(id=13, projectid=7, name='codebase1', slug='slug_codebase1'),
        ])

    def test_signature_find_codebase_id(self) -> None:
        @self.assertArgSpecMatches(
            self.master.data.updates.find_codebase_id,
            self.rtype.find_codebase_id,
        )
        def find_codebase_id(
            self: Any,
            *,
            projectid: int,
            name: int,
            auto_create: bool = True,
        ) -> None:
            pass

    def test_signature_update_codebase_info(self) -> None:
        @self.assertArgSpecMatches(
            self.master.data.updates.update_codebase_info,
            self.rtype.update_codebase_info,
        )
        def update_codebase_info(
            self: Any,
            *,
            codebaseid: int,
            projectid: int,
            slug: str,
        ) -> None:
            pass

    @async_to_deferred
    async def test_update_find_codebase_id_exists(self) -> None:
        codebase_id = await self.master.data.updates.find_codebase_id(
            projectid=7,
            name='codebase1',
        )
        self.assertEqual(codebase_id, 13)

    @async_to_deferred
    async def test_update_find_codebase_id_new(self) -> None:
        codebase_id = await self.master.data.updates.find_codebase_id(
            projectid=8,
            name='codebase2',
        )
        codebase = await self.master.data.get(('codebases', codebase_id))
        self.assertEqual(
            codebase,
            {'codebaseid': codebase_id, 'name': 'codebase2', 'slug': 'codebase2', 'projectid': 8},
        )

    @async_to_deferred
    async def test_update_codebase_info(self) -> None:
        await self.master.data.updates.update_codebase_info(
            codebaseid=13,
            projectid=8,
            slug='new_slug',
        )
        codebases = await self.master.data.get(('codebases', 13))
        self.assertEqual(
            codebases,
            {'codebaseid': 13, 'name': 'codebase1', 'slug': 'new_slug', 'projectid': 8},
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from typing import ClassVar
from typing import Sequence
from unittest import mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.db.schedulers import SchedulerModel
from buildbot.schedulers import base
from buildbot.schedulers import manager
from buildbot.test.util.warnings import assertProducesWarnings
from buildbot.warnings import DeprecatedApiWarning


class SchedulerManager(unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.next_objectid = 13
        self.objectids = {}

        self.master = mock.Mock()
        self.master.master = self.master

        def getObjectId(sched_name, class_name):
            k = (sched_name, class_name)
            try:
                rv = self.objectids[k]
            except KeyError:
                rv = self.objectids[k] = self.next_objectid
                self.next_objectid += 1
            return defer.succeed(rv)

        self.master.db.state.getObjectId = getObjectId

        def getScheduler(sched_id):
            return defer.succeed(
                SchedulerModel(
                    id=sched_id,
                    name=f"test-{sched_id}",
                    enabled=True,
                    masterid=None,
                )
            )

        self.master.db.schedulers.getScheduler = getScheduler

        self.new_config = mock.Mock()

        self.sm = manager.SchedulerManager()
        yield self.sm.setServiceParent(self.master)
        yield self.sm.startService()

    def tearDown(self):
        if self.sm.running:
            return self.sm.stopService()
        return None

    class Sched(base.BaseScheduler):
        # changing sch.attr should make a scheduler look "updated"
        compare_attrs: ClassVar[Sequence[str]] = ('attr',)
        already_started = False
        reconfig_count = 0

        def startService(self):
            assert not self.already_started
            assert self.master is not None
            assert self.objectid is not None
            self.already_started = True
            return super().startService()

        @defer.inlineCallbacks
        def stopService(self):
            yield super().stopService()

            assert self.master is not None
            assert self.objectid is not None

        def __repr__(self):
            return f"{self.__class__.__name__}(attr={self.attr})"

    class ReconfigSched(Sched):
        def reconfigServiceWithSibling(self, sibling):
            self.reconfig_count += 1
            self.attr = sibling.attr
            return super().reconfigServiceWithSibling(sibling)

    class ReconfigSched2(ReconfigSched):
        pass

    def makeSched(self, cls, name, attr='alpha'):
        with assertProducesWarnings(
            DeprecatedApiWarning, message_pattern='.*BaseScheduler has been deprecated.*'
        ):
            sch = cls(name=name, builderNames=['x'], properties={})
        sch.attr = attr
        return sch

    # tests

    @defer.inlineCallbacks
    def test_reconfigService_add_and_change_and_remove(self):
        sch1 = self.makeSched(self.ReconfigSched, 'sch1', attr='alpha')
        self.new_config.schedulers = {"sch1": sch1}

        yield self.sm.reconfigServiceWithBuildbotConfig(self.new_config)

        self.assertIdentical(sch1.parent, self.sm)
        self.assertIdentical(sch1.master, self.master)
        self.assertEqual(sch1.reconfig_count, 1)

        sch1_new = self.makeSched(self.ReconfigSched, 'sch1', attr='beta')
        sch2 = self.makeSched(self.ReconfigSched, 'sch2', attr='alpha')
        self.new_config.schedulers = {"sch1": sch1_new, "sch2": sch2}

        yield self.sm.reconfigServiceWithBuildbotConfig(self.new_config)

        # sch1 is still the active scheduler, and has been reconfig'd,
        # and has the correct attribute
        self.assertIdentical(sch1.parent, self.sm)
        self.assertIdentical(sch1.master, self.master)
        self.assertEqual(sch1.attr, 'beta')
        self.assertEqual(sch1.reconfig_count, 2)
        self.assertIdentical(sch1_new.parent, None)
        self.assertIdentical(sch1_new.master, None)

        self.assertIdentical(sch2.parent, self.sm)
        self.assertIdentical(sch2.master, self.master)

        self.new_config.schedulers = {}

        self.assertEqual(sch1.running, True)
        yield self.sm.reconfigServiceWithBuildbotConfig(self.new_config)
        self.assertEqual(sch1.running, False)

    @defer.inlineCallbacks
    def test_reconfigService_class_name_change(self):
        sch1 = self.makeSched(self.ReconfigSched, 'sch1')
        self.new_config.schedulers = {"sch1": sch1}

        yield self.sm.reconfigServiceWithBuildbotConfig(self.new_config)

        self.assertIdentical(sch1.parent, self.sm)
        self.assertIdentical(sch1.master, self.master)
        self.assertEqual(sch1.reconfig_count, 1)

        sch1_new = self.makeSched(self.ReconfigSched2, 'sch1')
        self.new_config.schedulers = {"sch1": sch1_new}

        yield self.sm.reconfigServiceWithBuildbotConfig(self.new_config)

        # sch1 had its class name change, so sch1_new is now the active
        # instance
        self.assertIdentical(sch1_new.parent, self.sm)
        self.assertIdentical(sch1_new.master, self.master)

    @defer.inlineCallbacks
    def test_reconfigService_not_reconfigurable(self):
        sch1 = self.makeSched(self.Sched, 'sch1', attr='beta')
        self.new_config.schedulers = {"sch1": sch1}

        yield self.sm.reconfigServiceWithBuildbotConfig(self.new_config)

        self.assertIdentical(sch1.parent, self.sm)
        self.assertIdentical(sch1.master, self.master)

        sch1_new = self.makeSched(self.Sched, 'sch1', attr='alpha')
        self.new_config.schedulers = {"sch1": sch1_new}

        with assertProducesWarnings(
            DeprecatedApiWarning, message_pattern='.*raising NotImplementedError.*'
        ):
            yield self.sm.reconfigServiceWithBuildbotConfig(self.new_config)

        # sch1 had parameter change but is not reconfigurable, so sch1_new is now the active
        # instance
        self.assertEqual(sch1_new.running, True)
        self.assertEqual(sch1.running, False)
        self.assertIdentical(sch1_new.parent, self.sm)
        self.assertIdentical(sch1_new.master, self.master)

    @defer.inlineCallbacks
    def test_reconfigService_not_reconfigurable_no_change(self):
        sch1 = self.makeSched(self.Sched, 'sch1', attr='beta')
        self.new_config.schedulers = {"sch1": sch1}

        yield self.sm.reconfigServiceWithBuildbotConfig(self.new_config)

        self.assertIdentical(sch1.parent, self.sm)
        self.assertIdentical(sch1.master, self.master)

        sch1_new = self.makeSched(self.Sched, 'sch1', attr='beta')
        self.new_config.schedulers = {"sch1": sch1_new}

        yield self.sm.reconfigServiceWithBuildbotConfig(self.new_config)

        # sch1 had its class name change, so sch1_new is now the active
        # instance
        self.assertIdentical(sch1_new.parent, None)
        self.assertEqual(sch1_new.running, False)
        self.assertIdentical(sch1_new.master, None)
        self.assertEqual(sch1.running, True)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from parameterized import parameterized
from twisted.internet import defer
from twisted.trial import unittest

from buildbot import config
from buildbot.process.results import FAILURE
from buildbot.process.results import SUCCESS
from buildbot.process.results import WARNINGS
from buildbot.steps import python
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import ExpectShell
from buildbot.test.steps import TestBuildStepMixin

log_output_success = """\
Making output directory...
Running Sphinx v1.0.7
loading pickled environment... not yet created
No builder selected, using default: html
building [html]: targets for 24 source files that are out of date
updating environment: 24 added, 0 changed, 0 removed
reading sources... [  4%] index
reading sources... [  8%] manual/cfg-builders
...
copying static files... done
dumping search index... done
dumping object inventory... done
build succeeded.
"""

log_output_nochange = """\
Running Sphinx v1.0.7
loading pickled environment... done
No builder selected, using default: html
building [html]: targets for 0 source files that are out of date
updating environment: 0 added, 0 changed, 0 removed
looking for now-outdated files... none found
no targets are out of date.
"""

log_output_warnings = """\
Running Sphinx v1.0.7
loading pickled environment... done
building [html]: targets for 1 source files that are out of date
updating environment: 0 added, 1 changed, 0 removed
reading sources... [100%] file

file.rst:18: (WARNING/2) Literal block expected; none found.

looking for now-outdated files... none found
pickling environment... done
checking consistency... done
preparing documents... done
writing output... [ 50%] index
writing output... [100%] file

index.rst:: WARNING: toctree contains reference to document 'preamble' that \
doesn't have a title: no link will be generated
writing additional files... search
copying static files... done
dumping search index... done
dumping object inventory... done
build succeeded, 2 warnings."""

log_output_warnings_strict = """\
Running Sphinx v1.0.7
loading pickled environment... done
building [html]: targets for 1 source files that are out of date
updating environment: 0 added, 1 changed, 0 removed
reading sources... [100%] file

Warning, treated as error:
file.rst:18:Literal block expected; none found.
"""

warnings = """\
file.rst:18: (WARNING/2) Literal block expected; none found.
index.rst:: WARNING: toctree contains reference to document 'preamble' that \
doesn't have a title: no link will be generated\
"""

# this is from a run of epydoc against the buildbot source..
epydoc_output = """\
  [...............
+---------------------------------------------------------------------
| In /home/dustin/code/buildbot/t/buildbot/master/buildbot/
| ec2.py:
| Import failed (but source code parsing was successful).
|     Error: ImportError: No module named boto (line 19)
|
  [....
Warning: Unable to extract the base list for
         twisted.web.resource.EncodingResourceWrapper: Bad dotted name
  [......
+---------------------------------------------------------------------
| In /home/dustin/code/buildbot/t/buildbot/master/buildbot/worker/
| ec2.py:
| Import failed (but source code parsing was successful).
|     Error: ImportError: No module named boto (line 28)
|
  [...........
+---------------------------------------------------------------------
| In /home/dustin/code/buildbot/t/buildbot/master/buildbot/status/
| status_push.py:
| Import failed (but source code parsing was successful).
|     Error: ImportError: No module named status_json (line 40)
|
  [....................<paragraph>Special descriptor for class __provides__</paragraph>
"""


class BuildEPYDoc(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def test_sample(self):
        self.setup_step(python.BuildEPYDoc())
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['make', 'epydocs']).stdout(epydoc_output).exit(1),
        )
        self.expect_outcome(result=FAILURE, state_string='epydoc warn=1 err=3 (failure)')
        return self.run_step()


class PyLint(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    @parameterized.expand([('no_results', True), ('with_results', False)])
    def test_success(self, name, store_results):
        self.setup_step(python.PyLint(command=['pylint'], store_results=store_results))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['pylint'])
            .stdout('Your code has been rated at 10/10')
            .exit(python.PyLint.RC_OK)
        )
        self.expect_outcome(result=SUCCESS, state_string='pylint')
        if store_results:
            self.expect_test_result_sets([('Pylint warnings', 'code_issue', 'message')])
            self.expect_test_results([])
        return self.run_step()

    @parameterized.expand([('no_results', True), ('with_results', False)])
    def test_error(self, name, store_results):
        self.setup_step(python.PyLint(command=['pylint'], store_results=store_results))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['pylint'])
            .stdout(
                'W: 11: Bad indentation. Found 6 spaces, expected 4\n'
                'E: 12: Undefined variable \'foo\'\n'
            )
            .exit(python.PyLint.RC_WARNING | python.PyLint.RC_ERROR)
        )
        self.expect_outcome(result=FAILURE, state_string='pylint error=1 warning=1 (failure)')
        self.expect_property('pylint-warning', 1)
        self.expect_property('pylint-error', 1)
        if store_results:
            self.expect_test_result_sets([('Pylint warnings', 'code_issue', 'message')])
            # note that no results are submitted for tests where we don't know the location
        return self.run_step()

    def test_header_output(self):
        self.setup_step(python.PyLint(command=['pylint'], store_results=False))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['pylint'])
            .log('stdio', header='W: 11: Bad indentation. Found 6 spaces, expected 4\n')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string='pylint')
        return self.run_step()

    def test_failure(self):
        self.setup_step(python.PyLint(command=['pylint'], store_results=False))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['pylint'])
            .stdout(
                'W: 11: Bad indentation. Found 6 spaces, expected 4\n'
                'F: 13: something really strange happened\n'
            )
            .exit(python.PyLint.RC_WARNING | python.PyLint.RC_FATAL)
        )
        self.expect_outcome(result=FAILURE, state_string='pylint fatal=1 warning=1 (failure)')
        self.expect_property('pylint-warning', 1)
        self.expect_property('pylint-fatal', 1)
        return self.run_step()

    def test_failure_zero_returncode(self):
        # Make sure that errors result in a failed step when pylint's
        # return code is 0, e.g. when run through a wrapper script.
        self.setup_step(python.PyLint(command=['pylint'], store_results=False))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['pylint'])
            .stdout(
                'W: 11: Bad indentation. Found 6 spaces, expected 4\n'
                'E: 12: Undefined variable \'foo\'\n'
            )
            .exit(0)
        )
        self.expect_outcome(result=FAILURE, state_string='pylint error=1 warning=1 (failure)')
        self.expect_property('pylint-warning', 1)
        self.expect_property('pylint-error', 1)
        return self.run_step()

    def test_regex_text(self):
        self.setup_step(python.PyLint(command=['pylint'], store_results=False))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['pylint'])
            .stdout(
                'W: 11: Bad indentation. Found 6 spaces, expected 4\n'
                'C:  1:foo123: Missing docstring\n'
            )
            .exit(python.PyLint.RC_WARNING | python.PyLint.RC_CONVENTION)
        )
        self.expect_outcome(
            result=WARNINGS, state_string='pylint convention=1 warning=1 (warnings)'
        )
        self.expect_property('pylint-warning', 1)
        self.expect_property('pylint-convention', 1)
        self.expect_property('pylint-total', 2)
        return self.run_step()

    def test_regex_text_0_24(self):
        # pylint >= 0.24.0 prints out column offsets when using text format
        self.setup_step(python.PyLint(command=['pylint'], store_results=False))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['pylint'])
            .stdout(
                'W: 11,0: Bad indentation. Found 6 spaces, expected 4\n'
                'C:  3,10:foo123: Missing docstring\n'
            )
            .exit(python.PyLint.RC_WARNING | python.PyLint.RC_CONVENTION)
        )
        self.expect_outcome(
            result=WARNINGS, state_string='pylint convention=1 warning=1 (warnings)'
        )
        self.expect_property('pylint-warning', 1)
        self.expect_property('pylint-convention', 1)
        self.expect_property('pylint-total', 2)
        return self.run_step()

    def test_regex_text_1_3_1(self):
        # at least pylint 1.3.1 prints out space padded column offsets when
        # using text format
        self.setup_step(python.PyLint(command=['pylint'], store_results=False))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['pylint'])
            .stdout(
                'W: 11, 0: Bad indentation. Found 6 spaces, expected 4\n'
                'C:  3,10:foo123: Missing docstring\n'
            )
            .exit(python.PyLint.RC_WARNING | python.PyLint.RC_CONVENTION)
        )
        self.expect_outcome(
            result=WARNINGS, state_string='pylint convention=1 warning=1 (warnings)'
        )
        self.expect_property('pylint-warning', 1)
        self.expect_property('pylint-convention', 1)
        self.expect_property('pylint-total', 2)
        return self.run_step()

    @parameterized.expand([('no_results', True), ('with_results', False)])
    def test_regex_text_2_0_0(self, name, store_results):
        # pylint 2.0.0 changed default format to include file path
        self.setup_step(python.PyLint(command=['pylint'], store_results=store_results))

        stdout = (
            'test.py:9:4: W0311: Bad indentation. Found 6 spaces, expected 4 (bad-indentation)\n'
            + 'test.py:1:0: C0114: Missing module docstring (missing-module-docstring)\n'
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['pylint'])
            .stdout(stdout)
            .exit(python.PyLint.RC_WARNING | python.PyLint.RC_CONVENTION)
        )
        self.expect_outcome(
            result=WARNINGS, state_string='pylint convention=1 warning=1 (warnings)'
        )
        self.expect_property('pylint-warning', 1)
        self.expect_property('pylint-convention', 1)
        self.expect_property('pylint-total', 2)
        if store_results:
            self.expect_test_result_sets([('Pylint warnings', 'code_issue', 'message')])
            self.expect_test_results([
                (
                    1000,
                    'test.py:9:4: W0311: Bad indentation. Found 6 spaces, expected 4 '
                    + '(bad-indentation)',
                    None,
                    'test.py',
                    9,
                    None,
                ),
                (
                    1000,
                    'test.py:1:0: C0114: Missing module docstring (missing-module-docstring)',
                    None,
                    'test.py',
                    1,
                    None,
                ),
            ])
        return self.run_step()

    def test_regex_text_2_0_0_invalid_line(self):
        self.setup_step(python.PyLint(command=['pylint'], store_results=False))

        stdout = 'test.py:abc:0: C0114: Missing module docstring (missing-module-docstring)\n'

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['pylint'])
            .stdout(stdout)
            .exit(python.PyLint.RC_CONVENTION)
        )
        self.expect_outcome(result=SUCCESS, state_string='pylint')
        self.expect_property('pylint-warning', 0)
        self.expect_property('pylint-convention', 0)
        self.expect_property('pylint-total', 0)
        return self.run_step()

    def test_regex_text_ids(self):
        self.setup_step(python.PyLint(command=['pylint'], store_results=False))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['pylint'])
            .stdout('W0311: 11: Bad indentation.\nC0111:  1:funcName: Missing docstring\n')
            .exit(python.PyLint.RC_WARNING | python.PyLint.RC_CONVENTION)
        )
        self.expect_outcome(
            result=WARNINGS, state_string='pylint convention=1 warning=1 (warnings)'
        )
        self.expect_property('pylint-warning', 1)
        self.expect_property('pylint-convention', 1)
        self.expect_property('pylint-total', 2)
        return self.run_step()

    def test_regex_text_ids_0_24(self):
        # pylint >= 0.24.0 prints out column offsets when using text format
        self.setup_step(python.PyLint(command=['pylint'], store_results=False))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['pylint'])
            .stdout('W0311: 11,0: Bad indentation.\nC0111:  3,10:foo123: Missing docstring\n')
            .exit(python.PyLint.RC_WARNING | python.PyLint.RC_CONVENTION)
        )
        self.expect_outcome(
            result=WARNINGS, state_string='pylint convention=1 warning=1 (warnings)'
        )
        self.expect_property('pylint-warning', 1)
        self.expect_property('pylint-convention', 1)
        self.expect_property('pylint-total', 2)
        return self.run_step()

    @parameterized.expand([('no_results', True), ('with_results', False)])
    def test_regex_parseable_ids(self, name, store_results):
        self.setup_step(python.PyLint(command=['pylint'], store_results=store_results))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['pylint'])
            .stdout(
                'test.py:9: [W0311] Bad indentation.\n'
                'test.py:3: [C0111, foo123] Missing docstring\n'
            )
            .exit(python.PyLint.RC_WARNING | python.PyLint.RC_CONVENTION)
        )
        self.expect_outcome(
            result=WARNINGS, state_string='pylint convention=1 warning=1 (warnings)'
        )
        self.expect_property('pylint-warning', 1)
        self.expect_property('pylint-convention', 1)
        self.expect_property('pylint-total', 2)
        if store_results:
            self.expect_test_result_sets([('Pylint warnings', 'code_issue', 'message')])
            self.expect_test_results([
                (1000, 'test.py:9: [W0311] Bad indentation.', None, 'test.py', 9, None),
                (1000, 'test.py:3: [C0111, foo123] Missing docstring', None, 'test.py', 3, None),
            ])
        return self.run_step()

    def test_regex_parseable(self):
        self.setup_step(python.PyLint(command=['pylint'], store_results=False))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['pylint'])
            .stdout('test.py:9: [W] Bad indentation.\ntest.py:3: [C, foo123] Missing docstring\n')
            .exit(python.PyLint.RC_WARNING | python.PyLint.RC_CONVENTION)
        )
        self.expect_outcome(
            result=WARNINGS, state_string='pylint convention=1 warning=1 (warnings)'
        )
        self.expect_property('pylint-warning', 1)
        self.expect_property('pylint-convention', 1)
        self.expect_property('pylint-total', 2)
        return self.run_step()

    def test_regex_parseable_1_3_1(self):
        """In pylint 1.3.1, output parseable is deprecated, but looks like
        that, this is also the new recommended format string:
            --msg-template={path}:{line}: [{msg_id}({symbol}), {obj}] {msg}
        """
        self.setup_step(python.PyLint(command=['pylint'], store_results=False))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['pylint'])
            .stdout(
                'test.py:9: [W0311(bad-indentation), ] '
                'Bad indentation. Found 6 '
                'spaces, expected 4\n'
                'test.py:3: [C0111(missing-docstring), myFunc] Missing '
                'function docstring\n'
            )
            .exit(python.PyLint.RC_WARNING | python.PyLint.RC_CONVENTION)
        )
        self.expect_outcome(
            result=WARNINGS, state_string='pylint convention=1 warning=1 (warnings)'
        )
        self.expect_property('pylint-warning', 1)
        self.expect_property('pylint-convention', 1)
        self.expect_property('pylint-total', 2)
        return self.run_step()


class PyFlakes(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def test_success(self):
        self.setup_step(python.PyFlakes())
        self.expect_commands(ExpectShell(workdir='wkdir', command=['make', 'pyflakes']).exit(0))
        self.expect_outcome(result=SUCCESS, state_string='pyflakes')
        return self.run_step()

    def test_content_in_header(self):
        self.setup_step(python.PyFlakes())
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['make', 'pyflakes'])
            # don't match pyflakes-like output in the header
            .log('stdio', header="foo.py:1: 'bar' imported but unused\n")
            .exit(0)
        )
        self.expect_outcome(result=0, state_string='pyflakes')
        return self.run_step()

    def test_unused(self):
        self.setup_step(python.PyFlakes())
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['make', 'pyflakes'])
            .stdout("foo.py:1: 'bar' imported but unused\n")
            .exit(1)
        )
        self.expect_outcome(result=WARNINGS, state_string='pyflakes unused=1 (warnings)')
        self.expect_property('pyflakes-unused', 1)
        self.expect_property('pyflakes-total', 1)
        return self.run_step()

    def test_undefined(self):
        self.setup_step(python.PyFlakes())
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['make', 'pyflakes'])
            .stdout("foo.py:1: undefined name 'bar'\n")
            .exit(1)
        )
        self.expect_outcome(result=FAILURE, state_string='pyflakes undefined=1 (failure)')
        self.expect_property('pyflakes-undefined', 1)
        self.expect_property('pyflakes-total', 1)
        return self.run_step()

    def test_redefs(self):
        self.setup_step(python.PyFlakes())
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['make', 'pyflakes'])
            .stdout("foo.py:2: redefinition of unused 'foo' from line 1\n")
            .exit(1)
        )
        self.expect_outcome(result=WARNINGS, state_string='pyflakes redefs=1 (warnings)')
        self.expect_property('pyflakes-redefs', 1)
        self.expect_property('pyflakes-total', 1)
        return self.run_step()

    def test_importstar(self):
        self.setup_step(python.PyFlakes())
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['make', 'pyflakes'])
            .stdout("foo.py:1: 'from module import *' used; unable to detect undefined names\n")
            .exit(1)
        )
        self.expect_outcome(result=WARNINGS, state_string='pyflakes import*=1 (warnings)')
        self.expect_property('pyflakes-import*', 1)
        self.expect_property('pyflakes-total', 1)
        return self.run_step()

    def test_misc(self):
        self.setup_step(python.PyFlakes())
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['make', 'pyflakes'])
            .stdout("foo.py:2: redefinition of function 'bar' from line 1\n")
            .exit(1)
        )
        self.expect_outcome(result=WARNINGS, state_string='pyflakes misc=1 (warnings)')
        self.expect_property('pyflakes-misc', 1)
        self.expect_property('pyflakes-total', 1)
        return self.run_step()


class TestSphinx(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def test_builddir_required(self):
        with self.assertRaises(config.ConfigErrors):
            python.Sphinx()

    def test_bad_mode(self):
        with self.assertRaises(config.ConfigErrors):
            python.Sphinx(sphinx_builddir="_build", mode="don't care")

    def test_success(self):
        self.setup_step(python.Sphinx(sphinx_builddir="_build"))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['sphinx-build', '.', '_build'])
            .stdout(log_output_success)
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="sphinx 0 warnings")
        return self.run_step()

    def test_failure(self):
        self.setup_step(python.Sphinx(sphinx_builddir="_build"))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['sphinx-build', '.', '_build'])
            .stdout('oh noes!')
            .exit(1)
        )
        self.expect_outcome(result=FAILURE, state_string="sphinx 0 warnings (failure)")
        return self.run_step()

    def test_strict_warnings(self):
        self.setup_step(python.Sphinx(sphinx_builddir="_build", strict_warnings=True))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['sphinx-build', '-W', '.', '_build'])
            .stdout(log_output_warnings_strict)
            .exit(1)
        )
        self.expect_outcome(result=FAILURE, state_string="sphinx 1 warnings (failure)")
        return self.run_step()

    def test_nochange(self):
        self.setup_step(python.Sphinx(sphinx_builddir="_build"))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['sphinx-build', '.', '_build'])
            .stdout(log_output_nochange)
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="sphinx 0 warnings")
        return self.run_step()

    @defer.inlineCallbacks
    def test_warnings(self):
        self.setup_step(python.Sphinx(sphinx_builddir="_build"))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['sphinx-build', '.', '_build'])
            .stdout(log_output_warnings)
            .exit(0)
        )
        self.expect_outcome(result=WARNINGS, state_string="sphinx 2 warnings (warnings)")
        self.expect_log_file("warnings", warnings)
        yield self.run_step()

        self.assertEqual(self.get_nth_step(0).statistics, {'warnings': 2})

    def test_constr_args(self):
        self.setup_step(
            python.Sphinx(
                sphinx_sourcedir='src',
                sphinx_builddir="bld",
                sphinx_builder='css',
                sphinx="/path/to/sphinx-build",
                tags=['a', 'b'],
                strict_warnings=True,
                defines={"empty": None, "t": True, "f": False, "s": 'str'},
                mode='full',
            )
        )
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command=[
                    '/path/to/sphinx-build',
                    '-b',
                    'css',
                    '-t',
                    'a',
                    '-t',
                    'b',
                    '-D',
                    'empty',
                    '-D',
                    'f=0',
                    '-D',
                    's=str',
                    '-D',
                    't=1',
                    '-E',
                    '-W',
                    'src',
                    'bld',
                ],
            )
            .stdout(log_output_success)
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="sphinx 0 warnings")
        return self.run_step()
